#/*
# * Copyright (c) 2023 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import os
import getopt
import re
from pathlib import Path
from pathlib import PurePath
from lopper import Lopper
from lopper import LopperFmt
from lopper.yaml import LopperJSON
from lopper.tree import LopperAction
from lopper.tree import LopperTree
from lopper.tree import LopperNode
from lopper.tree import LopperProp
import lopper
import lopper_lib
from itertools import chain
import json
import humanfriendly

from lopper.log import _init, _warning, _info, _error, _debug
import logging

def is_compat( node, compat_string_to_test ):
    if re.search( "isospec,isospec-v1", compat_string_to_test):
        return isospec_domain
    if re.search( "module,isospec", compat_string_to_test):
        return isospec_domain
    return ""

# tests for a bit that is set, going fro 31 -> 0 from MSB to LSB
def check_bit_set(n, k):
    if n & (1 << (k)):
        return True

    return False

def set_bit(value, bit):
    return value | (1<<bit)

def clear_bit(value, bit):
    return value & ~(1<<bit)

def debug_exit( message = None ):
    if message:
        _info( message )
    _info( "debug exit" )
    os._exit(1)

class domain_yaml(object):

    # static / class viriable
    iso_cpus_to_device_tree_map = {
                                "APU*": {
                                          "compatible": "arm,cortex-a72",
                                          "el": 3
                                        },
                                "RPU*": {
                                          "compatible": "arm,cortex-r5",
                                          "el": None
                                        }
                              }

    def __init__( self, sdt = None ):
        self.tree = LopperTree()
        self.tree.phandle_resolution = False
        self.sdt = sdt


    def subsystem_add( self, subsystem_name="default-subsystem", subsystem_id=0 ):
        subsystems_node = LopperNode( abspath=f"/domains/{subsystem_name}", name=subsystem_name )
        subsystems_node["compatible"] = "xilinx,subsystem"
        subsystems_node["id"] = subsystem_id
        subsystems_node.phandle_resolution = False

        self.tree = self.tree + subsystems_node

        return subsystems_node

    def node_add( self, name, parent ):
        new_node = LopperNode( name=name )
        try:
            parent + new_node
        except Exception as e:
            _warning( f"node {new_node} could not be added: {e} " )

        return new_node

    def domain_add( self, domain_name="default", parent_domain = None, id=0 ):

        if not parent_domain:
            domain_node = LopperNode( abspath=f"/domains/{domain_name}", name=domain_name )
            domain_node["compatible"] = "openamp,domain-v1"
            domain_node["id"] = id
            self.tree = self.tree + domain_node
        else:
            domain_node = LopperNode( name=domain_name )
            domain_node["compatible"] = "openamp,domain-v1"
            domain_node["id"] = id
            _debug( f"               adding domain '{domain_name}' parent: {parent_domain}" )
            parent_domain + domain_node

        return domain_node

    def cpu_map( self, cpu_name ):
        cpu_map = {}
        for n,dn in domain_yaml.iso_cpus_to_device_tree_map.items():
            if re.search( n, cpu_name ):
                cpu_map = dn

        return cpu_map

    def device_flags_map( self, device_name, access ):
        domain_flag_dict = {}

        if type(access) == dict:
            # _info( f"device_flags_map: {access}" )
            try:
                flags = access["flags"]
                for flag,value in flags.items():
                    if value:
                        domain_flag_dict[flag] = True
            except:
                return domain_flag_dict
        else:
            # try 1: is it a property ?
            flags = access.propval( "flags" )

            # try 2: is it a subnode ?
            if not flags[0]:
                for n in access.children():
                    if n.name == "flags":
                        for p in n:
                            flags.append( p )

            # map the flags to something domains.yaml can output
            # create a flags dictionary, so we can next it into the access
            # structure below, which will then be transformed into yaml later.
            for flag in flags:
                try:
                    if flag.value != '':
                        # if a flag is present, it means it was set to "true", it
                        # won't even be here in the false case.
                        domain_flag_dict[flag.name] = True
                except:
                    pass

        # _info( "isospec_device_flags: %s %s" % (device_name,domain_flag_dict) )

        return domain_flag_dict

    def cpu_add( self, domain_or_subsystem, cpu ):
        # this should wrap the addition of an access entry
        _info( f"cpu_add: {domain_or_subsystem}: {cpu}" )

        # is there an cpu list already in the yaml node ?
        # better than this: add a routine to the yaml node to
        # abstract an cpu addition (via device)
        try:
            cpu_list = domain_or_subsystem["cpus"]
        except:
            cpu_list = LopperProp( "cpus", -1, domain_or_subsystem, [] )
            cpu_list.phandle_resolution = False
            ## TODO: we shouldn't need to do this, as the node is passed
            ##       to the init. This is a lopper tree bug
            domain_or_subsystem + cpu_list

        cpu_map = self.cpu_map( cpu["name"] )
        if cpu_map:
            compat_string = cpu_map["compatible"]
            device_tree_compat = compat_string
            cpu_name = cpu["name"]
        else:
            _error( f"unrecognized cpu {cpu_name}" )

        # did we have a mapped compatible string in the device tree ?
        if device_tree_compat:
            # is there a number in the isospec name ? If so, that is our
            # mask, if not, we set the cpu mask to 0x3 (them all)
            m = re.match( r'.*?(\d+)', cpu_name )
            if m:
                cpu_number = m.group(1)
            else:
                cpu_number = -1

            # look in the device tree for a node that matches the
            # mapped compatible string
            compatible_nodes = self.sdt.tree.cnodes( device_tree_compat )
            if compatible_nodes:
                # we need to find the cluster name / label, that's the parent
                # of the matching nodes, any node will do, so we take the first
                cpu_cluster = compatible_nodes[0].parent
                if not cpu_cluster:
                    _warning( f"no cluster found for cpus, returning" )
                    return None

                # take the label if set, otherwise take the node name
                cluster_name = cpu_cluster.label if cpu_cluster.label else cpu_cluster.name

                # we have the name, now we need the cluster mask. If
                # there's a cpu number. Confirm that the node exists,
                # and set the bit. If there's no number, our mask is
                # 0xf
                cluster_mask = 0
                if cpu_number != -1:
                    for c in compatible_nodes:
                        if re.search( "cpu@" + cpu_number, c.name ):
                            cluster_mask = set_bit( cluster_mask, int(cpu_number) )
                            cluster_cpu_label = c.label
                else:
                    cluster_cpu_label = ''
                    cluster_mask = 0xf

                # cpu mode checks.
                #    secure
                #    el
                try:
                    cpu_flags = cpus_info["flags"]
                except:
                    cpu_flags = {}

                secure = False
                mode_mask = 0
                try:
                    secure_val = cpu_flags["secure"]
                    secure = secure_val
                except Exception as e:
                    pass

                try:
                    mode = cpu_flags["mode"]
                    if mode == "el":
                        mode_mask = set_bit( mode_mask, 0 )
                        mode_mask = set_bit( mode_mask, 1 )
                except:
                    # no passed mode, use the el level from the cpu_map
                    if cpu_map:
                        mode_mask = cpu_map["el"]

                if mode_mask:
                    cpu_entry = { "dev": cluster_name,    # remove before writing to yaml (if no roundtrip)
                                  "spec_name": cpu_name,  # remove before writing to yaml (if no roundtrip)
                                  "cluster" : cluster_name,
                                  "cluster_cpu" : cluster_cpu_label,
                                  "cpumask" : hex(cluster_mask),
                                  "mode" : { "secure": secure,
                                             "el": hex(mode_mask)
                                            }
                                 }
                else:
                    cpu_entry = { "dev": cluster_name,    # remove before writing to yaml (if no roundtrip)
                                  "spec_name": cpu_name,  # remove before writing to yaml (if no roundtrip)
                                  "cluster" : cluster_name,
                                  "cluster_cpu" : cluster_cpu_label,
                                  "cpumask" : hex(cluster_mask),
                                  "mode" : { "secure": secure }
                                 }

                cpu_list.value.append( cpu_entry )
        else:
            _warning( f"cpus entry {cpus_info[c]} has no device tree mapping" )

    def memory_add( self, domain_or_subsystem, memory ):
        _info( f"memory_add: {domain_or_subsystem}: {memory}" )

        debug = False
        try:
            # is it explicitly tagged as memory ?
            mem_dest_flag = memory["mem"]
            memory_dest = "memory@.*"

            # does it have a nodeid ? if it does, it is not just DRAM
            # tag it via the regex.
            try:
                nodeid = memory["nodeid"]
                memory_dest = isospec.memory_dest( memory["name"] )
                memory_type = isospec.memory_type( memory["name"] )
            except:
                memory_type = "memory"
        except:
            # if it isn't, we have a regex match to figure
            # out what type of memory it may be
            memory_dest = isospec.memory_dest( memory["name"] )
            memory_type = isospec.memory_type( memory["name"] )

        # if memory_type == "sram":
        #     _info( "debug: sram found" )
        # _info( f"memory type: {memory_type}" )

        # is there an memory list already in the yaml node ?
        try:
            memory_list = domain_or_subsystem[memory_type]
        except:
            memory_list = LopperProp( memory_type, -1, domain_or_subsystem, [] )
            memory_list.phandle_resolution = False
            ## TODO: we shouldn't need to do this, as the node is passed
            ##       to the init. This is a lopper tree bug
            domain_or_subsystem + memory_list

        try:
            memory_dest_address = int(memory["addr"],16)
            memory_dest_size = memory["size"]
        except:
            memory_dest_address = None
            memory_dest_size = None

        possible_mem_nodes = []
        if memory_type == "memory":
            # we have a node to lookup in the device tree
            try:
                # Q: here's the problem. For SRAM this is called because
                #    there wasn't an address match when we looked up the
                #    memory. BUT, there are memory nodes in the tree, so
                #    this will return options. The SRAM may fall into one
                #    of the ranges, so we use that start/end address which
                #    then updates in the yaml. But that means the start
                #    address, etc, are lost (as is the type).
                #
                # Q: should we only declare a match if the start address
                #    matches, versus just being in the start + size of
                #    memory ? Only for SRAM or for all types of memory ?
                #
                possible_mem_nodes = self.sdt.tree.nodes(memory_dest)
                if debug:
                    _info( f"possible mem nodes: {possible_mem_nodes}" )
            except Exception as e:
                _info( f"Exception looking for memory: {e}" )

        # if there's no possible device nodes, then we double
        # check the type mapping
        if not possible_mem_nodes:
            memory_dest = isospec.memory_dest( memory["name"] )
            memory_type = isospec.memory_type( memory["name"] )


        ## Note: when we start to consider the found memory nodes
        ##       from above, the "dev" value is going to change
        ##       the name to that, instead of the name from the spec.

        # Q: do we need to check if the memory is fully contained ?
        #    We really only need the address for now, since we don't
        #    care if this is fully contained in a memory range .. but
        #    we might in the future if we don't end up adjusting the
        #    device tree, or if there are multiple possible memory
        #    nodes, we could find the best fit.

        if memory_dest_address != None:
            memory_list.value.append( { "dev": memory["name"],       # remove before writing to yaml (if no roundtrip)
                                        "spec_name": memory["name"], # remove before writing to yaml (if no roundtrip)
                                        "start": hex(memory_dest_address),
                                        "size":  memory_dest_size
                                      }
                                    )

    def device_add( self, domain_or_subsystem, device, flags ):
        # this should wrap the addition of an access entry
        _info( f"device_add: {domain_or_subsystem}: {device}" )

        # is there an access list already in the yaml node ?
        # better than this: add a routine to the yaml node to
        # abstract an access addition (via device)
        try:
            access_list = domain_or_subsystem["access"]
        except:
            access_list = LopperProp( "access", -1, domain_or_subsystem, [] )
            access_list.phandle_resolution = False
            ## TODO: we shouldn't need to do this, as the node is passed
            ##       to the init. This is a lopper tree bug
            domain_or_subsystem + access_list

        try:
            address = device['addr']
            tnodes = self.sdt.tree.addr_node( address )
            if not tnodes:
                raise Exception( f"No node found for: {device}" )

            # we take the first node, as that will be the parent node
            # in any sets of nodes with a 1:1 mapping, and is likely
            # what we want.
            tnode = tnodes[0]

        except Exception as e:
            _info( f"Exception while looking up node by address: {e}" )
            ## TODO: this may mean it is memory, processing should go
            ##        to a dedicated routine
            return

        try:
            access_list.value.append(
                                      {
                                         "dev": tnode.name,
                                         "spec_name": device["name"],
                                         "label": tnode.label,
                                         "flags": flags
                                      }
                                    )
        except Exception as e:
            _info( f"problem adding access: {e}" )

        #_info( f"device_add: access list updated" )


class isospec(object):

    # If something appears in this map, it is a memory entry, and
    # we need to process it as such.
    #
    # This is a static variable!
    #
    iso_memory_device_map = {
            "DDR0" : ["memory", "memory@.*"],
            "OCM.*" : ["sram", None],
            ".*TCM.*" : ["sram", None]
    }

    def __init__( self, json_file = None ):
        self.json = None
        self.json_tree = None
        self.default_settings = {}
        self.__base_protection = {}
        self.subsystems = {}
        self.domains = {}

        self.device_dict = {}
        self.smid_dict = {}
        self.other_dict = {}

        self.permissive = False

        # trackers are indexed by subsystem or domain name, then by
        # device name
        self.trackers = {}
        if json_file:
            self.json_read( json_file )
            self.setup()


    @classmethod
    def memory_type( cls, name ):
        mem_found = None
        for n,v in isospec.iso_memory_device_map.items():
            if re.search( n, name ):
                mem_found = v

        if mem_found:
            return mem_found[0]

        return "memory"

    @classmethod
    def memory_dest( cls, name ):
        mem_found = None
        for n,v in isospec.iso_memory_device_map.items():
            if re.search( n, name ):
                mem_found = v

        if mem_found:
            return mem_found[1]

        return ""

    def json_read( self, json_file ):
        # convert the spec to a LopperTree for consistent manipulation
        self.json = LopperJSON( json=json_file )
        self.json_tree = self.json.to_tree()

    def setup( self ):
        if self.json_tree:

            # process the info section of the .iss file
            try:
                if self.json_tree["/info"]:
                    format = self.json_tree["/info"]["format"].value
                    if format == "shallow":
                        _error( "shallow format iss files are not currently supported", True )
            except Exception as e:
                _warning( "depth of .iss file could not be determined, output may be incorrect" )

            # this is the default settings for devices, it also creates
            # a refcount dictionary
            self.default_settings = self.isospec_subsystem( "/default_settings/subsystems/default" )
            # the below are the global list of devices
            self.device_dict, self.smid_dict, self.other_dict = self.device_collect()

            subsystem_list = self.json_tree["/design/subsystems"].children()

            # check if there is a base_protection design
            try:
                base_protection = self.json_tree["/design/base_protection"]
                _info( "gathering base protection devices/memory/settings" )

                # This may eventually just end up being added to teh "default_settings", but
                # for now, we handle it separately, i.e: subsystem_list.append( base_protection )
                self.__base_protection = self.isospec_subsystem( "/default_settings/base_protection" )

            except Exception as e:
                _info( f"exception processing base protection: {e}" )
                pass

            s = self.json_tree["/design/subsystems"]

            ## Build a dictionary of subsystems
            for subsystem in subsystem_list:
                self.subsystems[subsystem.name] = {}
                # _info( f"subsystem: {subsystem}" )
                try:
                    # does it have access ?
                    subsystem_access = subsystem["access"]
                    for a in subsystem_access:
                        access_dest = self.access_target( a )
                        try:
                            # TODO: these trackers aren't really needed, consider
                            #       dropping them. We have the dedicated "trackers"
                            self.subsystems[subsystem.name][access_dest["name"]] = {
                                                                                     "refcount": 0,
                                                                                     "access": access_dest
                                                                                    }
                        except Exception as e:
                            _info( f"Exception creating tracker: {e}" )
                            continue

                    try:
                        domain_parent = self.json_tree[subsystem.abs_path + "/domains"]
                        for d in domain_parent.children():
                            self.domains[d.name] = {}
                            # this allows quick access to the domains dictionaries, which are kept
                            # in a flat dictionary, also for easy access
                            self.subsystems[subsystem.name]["domain:" + d.name] = self.domains[d.name]

                            domain_access = d["access"]
                            for a in domain_access:
                                access_dest = self.access_target( a )
                                try:
                                    # if we have name collisions, this either needs to
                                    # be nested, or we go with sysystem.name:d.name for
                                    # dictionary index
                                    ## TODO: similarly,we don't likely need the tracker, but
                                    ##       we do look these up later by the dictionary and name
                                    self.domains[d.name][access_dest["name"]] = {
                                                                                   "refcount" : 0,
                                                                                   "subsystem": subsystem.name,
                                                                                   "access": access_dest
                                                                                }
                                except Exception as e:
                                    _info( f"exception while setting up domain tracking: {e}" )

                    except:
                        pass
                except:
                    pass

    def isospec_subsystem( self, name ):
        subsystem_dict = {}

        json_tree = self.json_tree

        try:
            subsystem_node = json_tree[name]
        except:
            _warning( f"no defaults found under: {name}" )

        try:
            default_access = subsystem_node["access"]
            #access_list = []
            for d in range(len(default_access)):
                element =  default_access[d]
                # access_list.append( default_access[d] )
                try:
                    dname = element["name"]
                    try:
                        dtype = element["type"]
                    except:
                        dtype = "device"

                    try:
                        destinations = element["destinations"]
                    except:
                        try:
                            destinations = element["SMIDs"]
                        except:
                            destinations = []

                    try:
                        flags = element["flags"]
                    except:
                        flags = {}

                    _info( f"[{name}/access] device found:" )
                    _info( f"    name: {dname}" )
                    _info( f"    type: {dtype}" )
                    _info( f"    destinations: {destinations}" )
                    _info( f"    flags: {flags}" )

                    subsystem_dict[dname] = {
                        "refcount" : 0,
                        "name":  dname,
                        "type":  dtype,
                        "dests": destinations,
                        "flags": flags,
                        "json": element
                        }
                except Exception as e:
                    _info( f"exception: {e}" )
        except:
            _warning( f"subsystem: {name} has no access" )

        return subsystem_dict

    def device_collect( self ):
        device_dict = {}
        smid_dict = {}
        other_dict = {}
        isospec_json_tree = self.json_tree

        _info( f"collecting all possible devices" )
        try:
            # /design can have destinations as well
            design = isospec_json_tree["/design"]
            # otherwise it is in the cells
            design_cells = isospec_json_tree["/design/cells"]
        except:
            _warning( "no design/cells found in isolation spec" )
            return device_dict

        cell_list = [ design ]
        cell_list.extend( design_cells.subnodes() )
        for cell in cell_list:
            try:
                dests = cell["destinations"]
                _debug( f"processing cell: {cell.name}" )
                _debug( f"           destinations {dests.abs_path} [{len(dests)}]" )
                for d in range(len(dests)):
                    dest = dests[d]
                    _debug( f"                dest: {dest}" )
                    # A device has to have a nodeid for us to consider it, since
                    # otherwise it can't be referenced. The exception to this is
                    # memory, since memory entries never have nodeids.
                    try:
                        nodeid = dest["nodeid"]
                        device_dict[dest["name"]] = {
                                                      "refcount": 0,
                                                      "dest": dest
                                                    }
                    except:
                        try:
                            is_it_mem = dest["mem"]
                            # _info( f"we found memory: {dest}" )
                        except:
                            ## We could do the second regex match on other devices
                            ## to see if they are memory. i.e. DDRxy ..
                            is_it_mem = False

                        # this may be controlled by a command line option
                        # in the future
                        skip_memory = False

                        ## Q: We need to decide if memory always shows up in the global
                        ##    device list, even without a nodeid. since we are only checking
                        ##    for nodeid on non-memory flagged entries below.
                        if is_it_mem:
                            if not skip_memory:
                                device_dict[dest["name"]] = {
                                    "refcount": 0,
                                    "dest": dest
                                }
                            else:
                                _debug( "                memory detected, but skip is set" )
                        else:
                            # no nodeid, skip
                            other_dict[dest["name"]] = {
                                    "refcount": 0,
                                    "dest": dest
                                }
                            _debug( f"                   ** destination '{dest}' device has no nodeid, adding to 'other'" )

                dests = cell["SMIDs"]
                _debug( f"           SMIDs {dests.abs_path} [{len(dests)}]" )
                for d in range(len(dests)):
                    dest = dests[d]
                    _debug( f"                dest: {dest}" )
                    try:
                        name = dest["name"]
                        smid_dict[name] = {
                            "refcount": 0,
                            # Could be renmaed to "dests" to match the subsystem tracker type
                            "dest": dest
                            }
                    except:
                        _debug( f"                    skipping dest {dest} ({e})" )

            except:
                    continue

        return device_dict, smid_dict, other_dict

    def access_type( self, access_entry ):
        # the default is "device" if the access entry isn't
        # carrying any type information
        try:
            access_type = access_entry["type"]
        except:
            access_type = "device"

        return access_type

    def access_target( self, access_entry ):
        access = {}

        # does it have "same_as_default" ? in that case
        # we look up further, otherwise return what was
        # passed in. This allows the caller to abstract
        # where there access definition comes from
        try:
            name = access_entry["same_as_default"]
            access = self.default_settings[name]["json"]
        except Exception as e:
            # try base protections
            try:
                access = self.__base_protection[name]["json"]
            except:
                # there's no entry, and no base protection, just
                # return what was passed in. This allows the
                # caller to normalize their access processing
                # by always getting something back
                access = access_entry

        return access

    def access_flags( self, access, translate=False ):
        flags = {}
        try:
            flags = access["flags"]
            # placeholder for flag translation / mapping, for
            # now we just return the raw flags
        except:
            pass

        return flags

    ## just returns the names, not the device
    def dests( self, access ):
        destlist = []
        try:
            try:
                destinations = access["destinations"]
                destlist.extend( destinations )
                # for now, we only allow one type of destination
                return destlist
            except:
                pass

            try:
                #_info( f"checking smids: {access}" )
                destinations = access["SMIDs"]
                destlist.extend( destinations )
                # for now, we only allow one type of destination
                return destlist
            except:
                pass
        except:
            pass

        return destlist

    # find a device by access, optionally getting the
    # flags as well
    def access_devices( self, access, return_flags=False ):
        destlist = []
        try:
            try:
                destinations = access["destinations"]
                dests = self.devices( destinations )
                destlist.extend( dests )
            except:
                pass

            try:
                #_info( f"checking smids: {access}" )
                destinations = access["SMIDs"]
                for d in destinations:
                    dests = self.devices( d )
                    destlist.extend( dests )
            except:
                pass
        except:
            pass

        return destlist

    ## find a device by name
    def devices( self, device_name_list, return_other = False ):
        devlist = []
        try:
            if self.device_dict:
                for d in device_name_list:
                    device = self.device_dict[d]
                    devlist.append( device['dest'] )
        except:
            pass

        if return_other and self.permissive:
            try:
                if self.other_dict:
                    for d in device_name_list:
                        device = self.other_dict[d]
                        devlist.append( device['dest'] )
            except:
                pass

        return devlist

    ## find a cpu by name
    ## TODO: we could just make devices() try this if it
    ##       fails in device lookup. That way we don't push
    ##       the type detection onto the caller
    def cpus( self, cpu_name_list ):
        cpulist = []
        try:
            if self.smid_dict:
                for c in cpu_name_list:
                    cpu = self.smid_dict[c]
                    cpulist.append( cpu['dest'] )
        except:
            pass

        return cpulist

    def is_subsystem( self, name ):
        sub = self.subsystem( name )
        if sub:
            return True

        return false

    def base_protection( self ):
        try:
            base_protection = self.json_tree["/design/base_protection"]
            return base_protection
        except:
            return None


    def subsystem( self, name = None ):
        """ returns all design subsystems, or a specific design
            subsystem if a name is passed
        """
        if not self.json_tree:
            return []

        try:
            nodes = self.json_tree["/design/subsystems"]
            if name:
                for n in nodes.children():
                    if n.name == name:
                        return [ n ]
                return []
            else:
                design_subs = nodes.children()
                # Note: if we want to process base protection as a standard
                #       subsystem, it could be looked up and added here
                #       like the following commented block:
                # try:
                #     base_protection = self.json_tree["/design/base_protection"]
                #     design_subs.append( base_protection )
                # except:
                #     pass
                return design_subs
        except:
            return []


    def subsystem_container( self, spec_node ):
        """ finds the subsystem that contains a given node
        """
        subsystem = None
        spec_parent_node = spec_node.parent
        while spec_parent_node:
            try:
                id = spec_parent_node["id"].value
                subsystem = spec_parent_node
            except Exception as e:
                # no id, it isn't a subsystem
                pass

            spec_parent_node = spec_parent_node.parent

        return subsystem

    def domain( self, subsystem, domain_name = None ):
        """ gets all the domains of a subsystem, if a name is
            passed, return a specific domain of the subsystem
        """
        try:
            # this could be a list comprehension
            domains = subsystem.children()
            if not domain_name:
                for d in domains:
                    return d.children()
            else:
                for d in domains:
                    for dd in d.children():
                        if dd.name == domain_name:
                            return [dd]

                return []
        except:
            return []

        return []

    def tracker_init( self, subsystem_name ):
        # The node indexed dictionary will be device names and a
        # True/False if it is referenced.
        self.trackers[subsystem_name] = {
                                          "mem" : {},
                                          "dev" : {},
                                          "cpu" : {}
                                        }

    def track_ref( self, tracker_name, dev, ttype = "dev", value = True ):
        ## when called with value as the default, we are initializing
        ## the tracking entry for a given name. Call with "true" to
        ## reference a device
        try:
            tracker = self.trackers[tracker_name][ttype]
        except Exception as e:
            _debug( f"No tracker is initialized for {tracker_name}, cannot track device {dev}: {e}" )
            debug_exit()
            return

        # _debug( f"tracking!!: {tracker_name} ... {dev}" )
        tracker[dev["name"]] = value

    def tracker_dref( self, tracker_name, dev, ttype = "dev" ):
        ## returns True if something is refereced, false otherwise
        ## "dev" is an access dictonary
        try:
            tracker = self.trackers[tracker_name][ttype]
        except Exception as e:
            return False

        try:
            # if there's an entry at all, it is being
            # tracked, if it has a reference it gets updated
            # to True
            return tracker[dev["name"]]
        except:
            return False

    def tracker_check( self, tracker_name, dev, ttype = "dev" ):
        ## returns True if something is being tracked, false otherwise
        ## "dev" is an access dictonary
        try:
            tracker = self.trackers[tracker_name][ttype]
        except Exception as e:
            return False

        try:
            # if there's an entry at all, it is being
            # tracked.
            entry = tracker[dev["name"]]
            return True
        except:
            return False

    def tracker_get( self, tracker_name, ttype="dev" ):
        if ttype:
            try:
                tracker = self.trackers[tracker_name][ttype]
                return tracker
            except:
                return {}
        else:
            try:
                trackers = self.trackers[tracker_name]
                return trackers
            except:
                return {}

    def isodomain_convert( self, spec_node, domains_tree ):
        """ converts an isolation spec domain to a yaml domain
        """

        ## TODO: we should add the subsystem name, since there's
        ##       no guarantee at all that the domain names are unique

        yaml_node = domains_tree.tree.nodes( spec_node.name )[0]

        containing_subsystem = self.subsystem_container( spec_node )

        base_protection = self.base_protection()
        base_access_list = []
        if base_protection:
            _info( "iso_domain_convert: checking base protection" )
            access_list = base_protection["access"]
            try:
                access_chunks = json.loads( access_list.value )
                for access in access_chunks:
                    access = self.access_target( access )
                    _info( f"    base proection access: ({type(access)} {access}" )
                    try:
                        smids = access["SMIDs"]
                        smid_all = False
                        if 'ANY' in smids:
                            smid_all = True
                        else:
                            # Only items tagged with ANY for the SMID
                            # get processed to be added to domains.
                            continue
                    except Exception as e:
                        continue

                    # This covers devices and memory, since they are split
                    # apart in the main processing loop.
                    access_type = self.access_type( access )
                    if access_type == "device":
                        _info( f"      base protection device: {access['name']}" )
                        try:
                            dests = self.dests( access )
                            _info( f"       device dests: {dests}" )
                            devices = self.devices( dests, True )
                            _info( f"       devices: {devices}" )

                            ## if there are devices, they should be added to all
                            ## domains.

                            base_access_list.append( access )

                        except:
                            pass
                    else:
                        _info( f"non-device SMID any detected ... {access}" )

            except Exception as e:
                _info( f"Exception during base protection checking: {e}" )

        # The node indexed dictionary will be device names and a
        # True/False if it is referenced.
        self.tracker_init( spec_node.name )

        try:
            _info( f"isodomain_convert: {spec_node.name}" )

            domain = spec_node

            try:
                id = spec_node["id"]
                yaml_node + deepcopy( id )
            except Exception as e:
                pass

            try:
                access_list = spec_node["access"]
                access_chunks = json.loads( access_list.value )

                if base_access_list:
                    # NOTE: it is an open question if we should apply the base protection to
                    #        the default domain. If we want to exclude them from the default
                    #        domain, it is a simple enough name check, so for now, we do add
                    #        them to the default domain.
                    _info( "isodomain_convert: adding base access list to domain access list" )
                    access_chunks.extend( base_access_list )

                # for access in access_list:
                for access in access_chunks:
                    _info( f"isodomain_convert: processing access entry: ({type(access)} {access}" )

                    access = self.access_target( access )
                    access_type = self.access_type( access )

                    if access_type == "device":
                        try:
                            dests = self.dests( access )
                            # _info( f"                     device dests: {dests}" )

                            all_devs = False
                            if access in base_access_list:
                                # for now, base_protection devices don't have SMIDs, so
                                # we need to look at the "other" dict to find the device
                                # target
                                all_devs = True

                            devices = self.devices( dests, all_devs )
                            # _info( f"                     devices: {devices}" )

                            ## The device might be memory. We need to
                            ## check if if has the "mem": "true" flag,
                            ## or if it matches a regex.
                            for dev in devices:
                                try:
                                    mem_flag = dev["mem"]
                                except:
                                    mem_flag = False

                                if mem_flag:
                                    # _info( f"                   device is memory: {dev}" )
                                    domains_tree.memory_add( yaml_node, dev )
                                    self.track_ref(spec_node.name, dev, "mem", False)
                                else:
                                    try:
                                        # if dev["name"] in self.tracker_get( spec_node.name, "dev" ).keys():
                                        if self.tracker_check( spec_node.name, dev, "dev" ):
                                            _info( f"   duplicated device detected: {dev['name']}, skipping it ... " )
                                            continue
                                    except Exception as e:
                                        pass

                                    # add the devices to the node
                                    flags = domains_tree.device_flags_map( dev, access )
                                    domains_tree.device_add( yaml_node, dev, flags )

                                    # TODO: track memory and cpus as well
                                    #
                                    # initialize ourself to False, any subdomains will toggle this
                                    # to true if they do reference it (the second call here)
                                    self.track_ref(spec_node.name, dev, "dev", False)
                                    if containing_subsystem:
                                        try:
                                            # this is the containing subsystem, track the reference
                                            # there
                                            self.track_ref(containing_subsystem.name,dev, "dev" )
                                        except:
                                            pass

                            try:
                                mem = yaml_node["memory"]
                                if len(mem) == 1:
                                    # force an empty entry if there's only one memory, since this
                                    # ensures that the yaml will be in list form. If we don't
                                    # do this, then assists down the pipeline have to deal with
                                    # either lists or yaml nodes
                                    mem.value.append( {} )
                            except Exception as e:
                                pass

                        except:
                            pass
                    elif access_type == "cpu_list":
                        try:
                            dests = self.dests( access )
                            _info( f"                     cpus dests: {dests}" )
                            cpus = self.cpus( dests )
                            # _info( f"                     cpus: {cpus}" )
                            _info( f"processing cpu list: {cpus}" )
                            for c in cpus:
                                # add the cpus to the node
                                domains_tree.cpu_add( yaml_node, c )
                                self.track_ref(spec_node.name, c, "cpu", False)

                            if len(cpus) == 1:
                                # force an empty entry if there's only one cpu, since this
                                # ensures that the yaml will be in list form. If we don't
                                # do this, then assists down the pipeline have to deal with
                                # either lists or yaml nodes
                                cpu_list = yaml_node["cpus"]
                                cpu_list.value.append( {} )

                        except Exception as e:
                            _info( f"exception procesing cpus: {e}" )

                    elif access_type == "ss_management":
                        _info( f"spec type ss_management: {access}" )
                        _info( f"no action required, skipping" )
                    elif access_type == "ss_permissions":
                        _info( f"spec type ss_permissions: {access}" )
                        _info( f"no action required, skipping" )
                    else:
                        _error( f"unknown spec type: {access_type}" )

            except Exception as e:
                # no access
                # _info( f"no access found: {e}" )
                pass

        except Exception as e:
            _info( f"isosdomain_convert: no domain '{domain_name}' found" )

## kept for reference until SRAM is fixed
# def isospec_process_memory( name, dest, sdt, json_tree, debug = False ):
#     _info( f"isospec_process_memory: {dest}" )

#     try:
#         # is it explicitly tagged as memory ?
#         mem_dest_flag = dest["mem"]
#         memory_dest = "memory@.*"
#         memory_type = "memory"
#     except:
#         # if it isn't, we have a regex match to figure
#         # out what type of memory it may be
#         memory_dest = isospec_memory_dest( name )
#         memory_type = isospec_memory_type( name )
#         if memory_type == "sram":
#             os._exit(1)

#     possible_mem_nodes = []
#     if memory_type == "memory":
#         # we have a node to lookup in the device tree
#         try:
#             # Q: here's the problem. For SRAM this is called because
#             #    there wasn't an address match when we looked up the
#             #    memory. BUT, there are memory nodes in the tree, so
#             #    this will return options. The SRAM may fall into one
#             #    of the ranges, so we use that start/end address which
#             #    then updates in the yaml. But that means the start
#             #    address, etc, are lost (as is the type).
#             #
#             # Q: should we only declare a match if the start address
#             #    matches, versus just being in the start + size of
#             #    memory ? Only for SRAM or for all types of memory ?
#             #
#             possible_mem_nodes = sdt.tree.nodes(memory_dest)
#             if debug:
#                 _info( f"possible mem nodes: {possible_mem_nodes}" )
#         except Exception as e:
#             _info( f"Exception looking for memory: {e}" )

#     # if there's no possible device nodes, then we double
#     # check the type mapping
#     if not possible_mem_nodes:
#         memory_dest = isospec_memory_dest( name )
#         memory_type = isospec_memory_type( name )

#     # Q: do we need to check if it is fully contained ?
#     # We really only need the address for now, since we don't care if
#     # this is fully contained in a memory range .. but we might in the
#     # future if we don't end up adjusting the device tree, or if there
#     # are multiple possible memory nodes, we could find the best fit.
#     dest_start = int(dest['addr'],16)
#     dest_size = dest['size']

#     memory_node = None
#     memory_list = []
#     if memory_type == "memory":
#         _info( f"  memory: {memory_dest}" )

#         ##
#         ## This may no longer be correct. But this is looking at the
#         ## isospec memory, and seeing which system device tree nodes
#         ## it may fall into. Those nodes are then used to create
#         ## entries in the domains.yaml based on what we return here.
#         ##
#         ## If we are trying to adjust the SDT based on what is in the
#         ## isospec, then this maybe not be correct. We need to
#         ## clarify.
#         ##
#         memory_node_found=False
#         for n in possible_mem_nodes:
#             _info( f"  possible_mem_nodes: {n.abs_path} type: {n['device_type']}" )
#             try:
#                 if "memory" in n["device_type"].value:
#                     reg = n["reg"]
#                     _info( f"    reg {reg.value}" )

#                     # we could do this more generically and look it up
#                     # in the parent, but 2 is the default, so doing
#                     # this for initial effort
#                     address_cells = 2
#                     size_cells = 2

#                     reg_chunks = lopper_lib.chunks( reg.value, address_cells + size_cells )
#                     for reg_chunk in reg_chunks:
#                         start = reg_chunk[0:address_cells]
#                         start = lopper.base.lopper_base.encode_byte_array( start )
#                         start = int.from_bytes(start,"big")

#                         size =  reg_chunk[address_cells:]

#                         size = lopper.base.lopper_base.encode_byte_array( size )
#                         size = int.from_bytes(size,"big")

#                         _info( f"    start: {hex(start)} size: {hex(size)}" )

#                         ##
#                         ## Q: Should we be checking if our address
#                         ##    falls into this range ? or should be be
#                         ##    checking if the range should be adjusted
#                         ##    ? Something else ?
#                         ##
#                         ## Checking the range seems correct, and then
#                         ## when we add this to domains.yaml, it will
#                         ## adjust the device during final processing
#                         ##
#                         ## Without this, we get multiple mem entries
#                         ## per isospec target, and that is not useful
#                         ##
#                         ## Q: Should the start/size be the memory
#                         ##    start/size from the device tree, or from
#                         ##    the isospec ?  if they aren't from the
#                         ##    isospec, we don't have the information
#                         ##    to adjust the output devie tree.
#                         ##
#                         if dest_start >= start and dest_start <= start + size:
#                             _info( f"    memory is in range, adding: {name}" )
#                             memory_node_found = True
#                             memory_list.append( { "dev": name,          # remove before writing to yaml (if no roundtrip)
#                                                   "spec_name": name,    # remove before writing to yaml (if no roundtrip)
#                                                   "start": hex(start),
#                                                   "size": hex(size)
#                                                  }
#                                                )

#             except Exception as e:
#                 _debug( f"Exception {e}" )

#         if not memory_node_found:
#             # we could create one to match if this is the case, but for now, we warn.
#             _warning( f"no memory node found that contains '{dest}'" )

#     elif memory_type == "sram":
#         # no memory dest
#         _info( f"sram memory type: {memory_dest}" )
#         address = dest['addr']
#         tnode = sdt.tree.addr_node( address )
#         if tnode:
#             # pull the start and size out of the device tree node
#             # don't have a device tree to test this yet
#             _warning( f"    target node {tnode.abs_path} found, but no processing is implemented" )
#         else:
#             size = dest['size']
#             # size = humanfriendly.parse_size( size, True )
#             start = address
#             _info( f"    sram start: {start} size: {size}" )
#             memory_list.append( {
#                                   "dev": name,        # remove before writing to yaml (if no roundtrip)
#                                   "spec_name": name,  # remove before writing to yaml (if no roundtrip)
#                                   "start": start,
#                                   "size": size
#                                 }
#                               )

#     return memory_list

def isospec_domain( tgt_node, sdt, options ):
    """assist entry point, called from lopper when a node is
       identified, or passed as a command line assist
    """
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    try:
        args = options['args']
    except:
        args = []

    lopper.log._init( __name__ )

    # --permissive means that non-SMID devices/memory will be consulted
    opts,args2 = getopt.getopt( args, "mpvh", [ "help", "audit", "verbose", "permissive", "nomemory" ] )

    if opts == [] and args2 == []:
        usage()
        sys.exit(1)

    audit = False
    memory = True
    permissive = False
    for o,a in opts:
        # print( "o: %s a: %s" % (o,a))
        if o in ('-m', "--nomemory" ):
            memory = False
        elif o in ('-v', "--verbose"):
            verbose = verbose + 1
        elif o in ('-c', "--compare"):
            compare_list.append( a )
        elif o in ('-p', "--permissive"):
            permissive = True
        elif o in ("--audit"):
            audit = True
        elif o in ('-h', "--help"):
            # usage()
            sys.exit(1)

    if verbose:
        lopper.log._level( logging.INFO, __name__ )
    if verbose > 1:
        lopper.log._level( logging.DEBUG, __name__ )
        #logging.getLogger().setLevel( level=logging.DEBUG )

    _info( f"cb: isospec_domain( {tgt_node}, {sdt}, {verbose} )" )

    if sdt.support_files:
        isospecf = sdt.support_files.pop()
    else:
        try:
            if not args2[0]:
                _error( "isospec: no isolation specification passed" )
            isospecf = args2.pop(0)
        except Exception as e:
            _error( f"isospec: no isolation specification passed: {e}", True  )

    domain_yaml_file = "domains.yaml"
    try:
        domain_yaml_file = args2.pop(0)
    except:
        pass

    try:
        iso_file = Path( isospecf )
        iso_file_abs = iso_file.resolve( True )
    except FileNotFoundError as e:
        _error( f"ispec file {isospecf} not found", True )

    # convert the spec to a LopperTree for consistent manipulation
    json_in = LopperJSON( json=iso_file_abs )
    json_tree = json_in.to_tree()

    spec = isospec( iso_file_abs )
    spec.permissive = permissive

    ## self test block
    # access_node = spec.json_tree["/design/subsystems/default"]["access"]
    # _info( f"access node type: {type(access_node)}" )
    # for a in access_node:
    #     _info( f"a: {a}" )
    #     access_dest = spec.access_target( a )
    #     _info( f"access details: {access_dest}" )
    #     try:
    #         pdests = access_dest["destinations"]
    #         devices = spec.devices(pdests)
    #         _info( f"   devices: {devices}" )
    #     except:
    #         # no destination
    #         pass
    #     dests = spec.access_devices( access_dest )
    #     _info( f"   access destination devices: {dests}" )
    #
    # _info( "\n\n\n\n" )
    #
    # d = spec.json_tree["/design/subsystems/default/domains"]
    # for domain in d.children():
    #     _info( f"domain: {domain}" )
    #     for a in domain["access"]:
    #         access_dest = spec.access_target( a )
    #         #_info( f"access details: {access_dest}" )
    #         try:
    #             pdests = access_dest["destinations"]
    #             devices = spec.devices(pdests)
    #             _info( f"   devices: {devices}" )
    #         except:
    #             # no destinations
    #             pass
    #
    # for d in spec.json_tree["/design/subsystems/default/domains"]:
    #    _info( f"d: {d}" )
    # access_node = spec.json_tree["/design/subsystems/default"]["domains"]
    #
    # for a in range(len(access_node)):
    #     access_entry = access_node[a]
    #     _info( f"a: {access_entry}" )
    #
    # _info( "\n\n" )
    # _info( f"subsystems: {spec.subsystem('default')}" )
    #
    # domains = spec.domain( spec.subsystem('default')[0] )
    # _info( f"domains: {domains}" )
    # apu_domain = spec.domain( spec.subsystem('default')[0], "APU" )
    # _info( f"apu: {apu_domain}" )
    #
    # iso_subsystems = json_tree["/design/subsystems"]
    # try:
    #     iso_domains = json_tree["/design/subsystems/" ]
    # except:
    #     pass
    #
    ## end self test

    # create our domains.yaml tree
    domains = domain_yaml( sdt )

    # process the subsystems in the spec
    for sub in spec.subsystem():
        _info( f"processing: subsystem: {type(sub)} {sub}" )
        try:
            sub_id = sub["id"].value
        except:
            sub_id = 0
        subsystem_yaml_node = domains.subsystem_add( sub.name, sub_id )
        spec.isodomain_convert( sub, domains )

        # does the subsystem have domains ? if so, process them
        sub_domains = spec.domain( sub )
        if sub_domains:
            domain_container_node = domains.node_add( "domains", subsystem_yaml_node )
            for d in sub_domains:
                _info( f"    adding domain: {d.name}" )
                domain_node = domains.domain_add( d.name, domain_container_node, sub["id"] )
                spec.isodomain_convert( d, domains )

    # check our refcounter(s)
    if audit:
        lopper.log._level( logging.INFO, __name__ )
        for sub in spec.subsystem():
            try:
                trackers = spec.tracker_get(sub.name,None)
                _info( "" )
                header = f"Audit for subsystem: {sub.abs_path}"
                header_underline = "=" * len(header)
                _info( header )
                _info( header_underline )
                unrefd = []
                refd = []
                for dev,flag in trackers["dev"].items():
                    if flag:
                        refd.append( dev )
                    else:
                        unrefd.append( dev )

                mrefd = []
                munrefd = []
                for mem,flag in trackers["mem"].items():
                    if flag:
                        mrefd.append( mem )
                    else:
                        munrefd.append( mem )

                _info( f"  referenced devices  : {refd}" )
                _info( f"  unreferenced devices: {unrefd}" )
                _info( f"  referenced memory   : {mrefd}" )
                _info( f"  unreferenced memory : {munrefd}" )
                _info( "" )

                for d in spec.domain( sub ):
                    header = f"    Audit for domain: {d.name}"
                    header_underline = "    " + "-" * (len(header) - 4)
                    _info( header )
                    _info( header_underline )
                    tracker = spec.tracker_get(d.name,None)
                    devs = list(tracker["dev"].keys())
                    mem = list(tracker["mem"].keys())
                    _info( f"      devices: {devs}" )
                    _info( f"      memory : {mem}" )
            except:
                pass

    # write the yaml tree
    _info( f"writing domain file: {domain_yaml_file}" )
    sdt.write( domains.tree, output_filename=domain_yaml_file, overwrite=True )

    return True




#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
from lopper.tree import LopperAction
import lopper
import lopper.log as log

log._init(__name__)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,grep", compat_string_to_test):
        return grep
    return ""

# tests for a bit that is set, going fro 31 -> 0 from MSB to LSB
def check_bit_set(n, k):
    if n & (1 << (k)):
        return True

    return False

# tgt_node: is the openamp domain node number
# sdt: is the system device tree
def grep( tgt_node, sdt, options ):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    try:
        args = options['args']
    except:
        args = []

    log._debug(f"cb: grep( {tgt_node}, {sdt}, {verbose}, {args} )")

    node_regex = ""
    tgt_regex = args[0]
    if len(args) == 2:
        node_regex = args[1]

    # for n in sdt.tree:
    #     print( "n: %s" % n )

    nodes = []
    try:
        nodes = sdt.tree.nodes(node_regex)
        lnodes = sdt.tree.lnodes(node_regex)
        nodes = nodes + lnodes
    except:
        log._error(f"grep: nodes {node_regex} not found")
        sys.exit(1)


    matches = {}
    for n in nodes:
        try:
            match = n[tgt_regex]
            #print( "match: %s" % match )
            matches[n.abs_path] = match
            #print( "matches is now: %s" % matches )
        except Exception as e:
            pass

    if matches:
        for m in matches.keys():
            print( f"{m}: {matches[m]}")
    else:
        print( f"{tgt_regex}: not found" )

    return True



#/*
# * Copyright (c) 2026 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Ricky Sun <ricky.sun@amd.com>
# *       Francisco Iglesias <francisco.iglesias@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */
import sys
import os
import re
import common_utils as utils
from baremetalconfig_xlnx import scan_reg_size

sys.path.append(os.path.dirname(__file__))

def is_compat(node, compat_string_to_test):
    if re.search(r"module,.*gen_qemu_mem_cfg", compat_string_to_test):
        return gen_qemu_mem_cfg_all
    return ""


def gen_qemu_mem_cfg_all(tgt_node, sdt, options):
    """
    Wrapper that generates both the QEMU and matching Vitis NoC SystemC model
    memory config outputs.

    Args:
        tgt_node: Target node number (typically root)
        sdt: System device tree object
        options: Dictionary containing command-line options and args
    """
    gen_qemu_mem_cfg(tgt_node, sdt, options)
    noc_memory_config(tgt_node, sdt, options)
    return True


def extract_noc_memory_regions(tgt_node, sdt):
    """
    Extract NOC memory regions from the SDT.

    Traverses the device tree, filters for NOC memory nodes, and extracts
    address/size pairs from their reg properties.

    Args:
        tgt_node: Target node number (typically root)
        sdt: System device tree object

    Returns:
        List of (address, size) tuples, sorted by address.
        Empty list if no NOC memory found.
    """
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()

    memory_regions = []

    for node in root_sub_nodes:
        try:
            device_type = node.propval("device_type")
            if device_type != ["memory"]:
                continue

            if not any("xlnx,axi-noc" in t for t in node.type):
                continue

            reg_value = node.propval("reg")
            if not reg_value:
                continue

            na = node.parent.propval("#address-cells")[0]
            if not na:
                na = 2
            ns = node.parent.propval("#size-cells")[0]
            if not ns:
                ns = 2

            cells_per_region = na + ns
            num_regions = len(reg_value) // cells_per_region

            for idx in range(num_regions):
                try:
                    address, size = scan_reg_size(node, reg_value, idx)

                    if size == 0:
                        continue

                    memory_regions.append((address, size))

                except Exception as e:
                    continue

        except Exception as e:
            continue

    memory_regions.sort(key=lambda x: x[0])
    return memory_regions


def gen_qemu_mem_cfg(tgt_node, sdt, options):
    """
    Generate memory.qemuboot.conf with QEMU dynamic memory instantiation args.

    Scans the SDT for NOC memory regions and produces a qemuboot.conf file
    with -device amd-ddr-memory arguments for each region above 2GB.
    Regions below 2GB (0x80000000) are skipped as QEMU handles this region.

    Output format:
        [config_bsp]
        qb_mem = -machine dynamic-mem=on -device amd-ddr-memory,address=0x...,size=0x...

    Args:
        tgt_node: Target node number (typically root)
        sdt: System device tree object
        options: Dictionary containing command-line options and args
    """
    if not options.get('outdir', {}):
        raise ValueError("providing an outdir is required")

    outdir = options['outdir']
    memory_regions = extract_noc_memory_regions(tgt_node, sdt)

    if not memory_regions:
        return

    device_args = []
    for address, size in memory_regions:
        if address < 0x80000000:
            continue
        arg = f"-device amd-ddr-memory,address={hex(address)},size={hex(size)}"
        device_args.append(arg)

    if not device_args:
        return

    qb_mem = "-machine dynamic-mem=on " + " ".join(device_args)

    output_file = os.path.join(outdir, "memory.qemuboot.conf")

    if os.path.isfile(output_file) and "-f" not in options.get('args', []):
        print(f"[WARNING] {output_file} already exists, use -f to overwrite")
        return

    with open(output_file, 'w') as fd:
        fd.write("[config_bsp]\n")
        fd.write(f"qb_mem = {qb_mem}\n")

    return True


def noc_memory_config(tgt_node, sdt, options):
    """
    Generate noc_memory_config.txt matching the QEMU memory configuration.

    Extracts NOC memory regions from the SDT and writes them in the format:
        qemu-memory-_ddr@<hex_addr>,<decimal_address>,<decimal_size>

    Args:
        tgt_node: Target node number (typically root)
        sdt: System device tree object
        options: Dictionary containing command-line options and args
    """
    if not options.get('outdir', {}):
        raise ValueError("providing an outdir is required")
    outdir = options['outdir']

    memory_regions = extract_noc_memory_regions(tgt_node, sdt)

    if not memory_regions:
        return

    output_file = os.path.join(outdir, "noc_memory_config.txt")

    if os.path.isfile(output_file) and "-f" not in options.get('args', []):
        print(f"[WARNING] {output_file} already exists, use -f to overwrite")
        return

    with open(output_file, 'w') as fd:
        for address, size in memory_regions:
            addr_str = f"0x{address:08x}" if address == 0 else f"0x{address:x}"
            line = f"qemu-memory-_ddr@{addr_str},{address},{size}\n"
            fd.write(line)

    return True



#/*
# * Copyright (c) 2025, Advanced Micro Devices, Inc. All rights reserved.
# *
# * Author:
# *       Nava kishore Manne <nava.kishore.manne@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import os
import getopt
import re
import subprocess
from pathlib import Path
from pathlib import PurePath
from lopper import Lopper
from lopper import LopperFmt
import lopper
from re import *
import yaml
import glob
from collections import OrderedDict

sys.path.append(os.path.dirname(__file__))
from baremetalconfig_xlnx import *

def is_compat( node, compat_string_to_test ):
    if re.search( "module,xlnx_overlay_dt", compat_string_to_test):
        return xlnx_generate_overlay_dt
    return ""

def infer_platform_from_sdt(sdt):
    """
    Infer the platform from the system device tree by parsing the family property.

    Args:
        sdt: System device tree object

    Returns:
        str: Inferred processor name or None if unable to determine
    """
    try:
        # Map family to processor names
        family_to_processor = {
            'Zynq': 'cortexa9',
            'ZynqMP': 'cortexa53',
            'Versal': 'cortexa72',
            'VersalNet': 'cortexa78',
            'Versal_2VE_2VM': 'cortexa78'
        }

        # Try to infer from the 'family' property at the root node
        family = sdt.tree['/'].propval('family')[0]
        if family in family_to_processor:
            processor = family_to_processor[family]
            print(f"Automatically inferred processor '{processor}' from family property '{family}'")
            return processor

        # Fallback: infer from CPU compatible strings
        for cpu_node in sdt.tree.nodes('/cpu.*@.*'):
            compatible = cpu_node.propval('compatible')[0]
            if 'cortex-a9' in compatible:
                print("Automatically inferred processor 'cortexa9' from CPU compatible string")
                return 'cortexa9'
            elif 'cortex-a53' in compatible:
                print("Automatically inferred processor 'cortexa53' from CPU compatible string")
                return 'cortexa53'
            elif 'cortex-a72' in compatible:
                print("Automatically inferred processor 'cortexa72' from CPU compatible string")
                return 'cortexa72'
            elif 'cortex-a78' in compatible:
                print("Automatically inferred processor 'cortexa78' from CPU compatible string")
                return 'cortexa78'

    except Exception as e:
        print(f"Warning: Failed to infer platform from system device tree: {e}")

    return None


def usage():
    print(f"Usage: lopper -O <output_dir> -f --enhanced <system-top.dts> <lopper-gen DT> -- xlnx_overlay_dt [<machine>] <config> <PL DTSI (SDT-gen)>")
    print("\nArguments:")
    print("  <output_dir>      - Directory where the lopper generated overlay `pl.dtsi` file will be stored")
    print("  <system-top.dts>  - Full system device tree source")
    print("  <lopper-gen DT>   - Lopper-generated system device tree file (excluding PL node)")
    print("  <machine>         - (Optional) Target architecture: cortex-a9 | cortex-a53 | cortex-a72 | cortex-a78")
    print("                      If not provided, will be automatically inferred from system device tree")
    print("  <config>          - Configuration type: full | segmented | dfx | external-fpga-config")
    print("  <PL DTSI>         - pl.dtsi file, generated by the SDT tool")

    print("\nExamples:")
    print(" With machine argument:")
    print(" lopper -O <output_dir>/ -f --enhanced <path_to_system_top>/system-top.dts <path_to_lopper_gen_dt>/lopper-gen.dts -- xlnx_overlay_dt <machine> <config> <path_to_pl_dtsi>/pl.dtsi")
    print("")
    print(" With automatic platform inference:")
    print(" lopper -O <output_dir>/ -f --enhanced <path_to_system_top>/system-top.dts <path_to_lopper_gen_dt>/lopper-gen.dts -- xlnx_overlay_dt <config> <path_to_pl_dtsi>/pl.dtsi")
    print("  --firmware-name=<name> - (Optional) Override the default firmware-name in the output")

def remove_node_ref(sdt, tgt_node, ref_node):
    prop_dict = ref_node.__props__.copy()
    match_label_list = []
    for node in sdt.tree[tgt_node].subnodes():
        matched_label = get_label(sdt, ref_node, node)
        if matched_label:
            match_label_list.append(matched_label)
    for prop,node1 in prop_dict.items():
        if prop not in match_label_list:
            sdt.tree['/' + ref_node.name].delete(prop)

"""
This API removes lines that contain `status = "okay";` from a DTS/DTSI file.

Args:
	filepath (str): Path to the input DTS/DTSI file.
"""

def remove_status_okay(filepath):
    with open(filepath, "r") as f:
        lines = f.readlines()

    modified_lines = [
        line for line in lines if 'status = "okay";' not in line.strip()
    ]

    with open(filepath, "w") as f:
        f.writelines(modified_lines)

match_list = ["v_tc", "v_smpte_uhdsdi_tx", "v_smpte_uhdsdi_rx", "v_hdmi_tx1", "v_hdmi_rx1",
	      "v_hdmi_rx", "v_hdmi_tx", "v_dp_tx", "v_dp_rx", "v_dp_rx1", "v_dp_tx1"]
"""
This API adds `status = "disabled";` after each line matching `xlnx,ip-name = "{name}";`
from the provided match_list. Preserves indentation.
Args:
	filepath (str): Path to the input DTS/DTSI file.
	match_list (List[str]): List of IP names to match.
"""
def add_status_disabled_after_ipname(filepath, match_list):
    with open(filepath, "r") as f:
        lines = f.readlines()

    modified_lines = []
    for line in lines:
        modified_lines.append(line)
        stripped_line = line.strip()

        for name in match_list:
            if stripped_line == f'xlnx,ip-name = "{name}";':
                indent = line[:len(line) - len(line.lstrip())]
                modified_lines.append(f'{indent}status = "disabled";\n')
                break

    with open(filepath, "w") as f:
        f.writelines(modified_lines)


"""
This API generates the overlay dts file by taking pl.dtsi
generated from DTG++.
Args:
    tgt_node: is the baremetal config top level domain node number
    sdt:      is the system device-tree
    options:  There are two valid invocation modes:
              Mode 1 (3 args): <machine> <config> <PL DTSI>
              Mode 2 (2 args): <config> <PL DTSI> (machine auto-inferred)

              <machine> - Target architecture: cortex-a9 | cortex-a53 | cortex-a72 | cortex-a78
              <config>  - Configuration type: full | segmented | dfx | external-fpga-config
              <PL DTSI> - pl.dtsi file, generated by the SDT tool
"""
def xlnx_generate_overlay_dt(tgt_node, sdt, options):
    try:
        # Separate positional arguments from optional arguments
        positional_args = []
        optional_args = []

        for arg in options['args']:
            if arg.startswith("--"):
                optional_args.append(arg)
            else:
                positional_args.append(arg)

        # Process optional arguments
        firmware_override = None
        for arg in optional_args:
            if arg.startswith("--firmware-name="):
                firmware_override = arg.split("=", 1)[1]

        # Determine invocation mode based on number of positional arguments
        num_args = len(positional_args)
        processor = None
        config = None
        input_file = None

        if num_args == 3:
            # Mode 1: machine, config, input_file
            processor = positional_args[0]
            config = positional_args[1].strip().lower()
            input_file = positional_args[2]

        elif num_args == 2:
            # Mode 2: config, input_file (auto-infer machine)
            config = positional_args[0].strip().lower()
            input_file = positional_args[1]

            # Auto-infer processor from system device tree
            processor = infer_platform_from_sdt(sdt)
            if processor is None:
                print("Error: Unable to automatically infer platform from system device tree.")
                print("Please provide the machine argument explicitly.")
                usage()
                return False
        else:
            print("Error: Invalid number of arguments.")
            print(f"Expected 2 or 3 positional arguments, got {num_args}")
            usage()
            return False

    except (IndexError, KeyError) as e:
        print(f"Error: Failed to parse arguments: {e}")
        usage()
        return False

    # Validate config value
    valid_configs = {"full", "segmented", "dfx", "external-fpga-config"}
    if config not in valid_configs:
        print(f"Error: Invalid config '{config}'. Expected one of: {valid_configs}.")
        return False

    print(f"Processor: {processor}, Config: {config}, Input: {input_file}")

    platform_map = {
        "cortexa9-zynq": "Zynq",
        "cortexa9_0": "Zynq",
        "cortexa9": "Zynq",
        "cortexa53-zynqmp": "ZynqMP",
        "psu_cortexa53_0": "ZynqMP",
        "cortexa53_0": "ZynqMP",
        "cortexa53": "ZynqMP",
        "cortexa72-versal": "Versal",
        "psv_cortexa72_0": "Versal",
        "cortexa72_0": "Versal",
        "cortexa72": "Versal",
        "psx_cortexa78_0": "VersalNet",
        "cortexa78_0": "VersalNet",
        "cortexa78": "VersalNet"
    }

    platform = platform_map.get(processor)
    if platform is None:
        print(f"Error: Unsupported processor '{processor}'.")
        print(f"Expected one of: {', '.join(platform_map.keys())}.")
        return False

    # Read input file
    with open(input_file, "r") as infile:
        lines = infile.readlines()

    print("Starting overlay generation...")

    # Move the amba_pl node to the overlay
    try:
        amba_node = sdt.tree["/amba_pl"]
    except KeyError:
        print("Error: /amba_pl node not found in the system device tree.")
        return False


    new_amba_node = amba_node()
    sdt.tree = sdt.tree - amba_node  # Remove from base tree

    # Remove "amba_pl" references from __symbols__ and aliases
    remove_node_ref(sdt, tgt_node, sdt.tree['/__symbols__'])
    remove_node_ref(sdt, tgt_node, sdt.tree['/aliases'])

    filtered_lines = lines[1:-1]  # Remove first and last lines

    # Lines to remove if at the top of &amba
    if platform in ["Zynq"]:
        lines_to_remove = [
            "ranges;",
            "compatible = \"simple-bus\";",
            "#address-cells = <1>;",
            "#size-cells = <1>;"
        ]
    else:
        lines_to_remove = [
            "ranges;",
            "compatible = \"simple-bus\";",
            "#address-cells = <2>;",
            "#size-cells = <2>;"
        ]

    modified_lines = []
    amba_block = []
    fpga_pr_nodes = []
    misc_clk_nodes = []
    clocking_nodes = []
    afi_nodes = []
    amba_nodes = []

    inside_fpga_pr = inside_misc_clk = inside_clocking = inside_afi = inside_amba = False
    amba_started = False
    firmware_name = None

    i = 0
    while i < len(filtered_lines):
        stripped_line = filtered_lines[i].strip()

        # Capture firmware-name
        if "firmware-name" in stripped_line:
            firmware_name = filtered_lines[i]
            i += 1
            continue

        if stripped_line.startswith("amba_pl"):
            inside_amba = True
            amba_started = True
            amba_block.append("&amba {\n")
            i += 1
            continue

        if inside_amba:
            # Remove unnecessary lines at the top
            if all(
                i + j < len(filtered_lines) and filtered_lines[i + j].strip() == lines_to_remove[j]
                for j in range(len(lines_to_remove))
            ):
                i += len(lines_to_remove)  # Skip these lines
                continue
            inside_amba = False  # Stop once a non-matching line is found

        # Handle FPGA_PR nodes
        if "fpga_PR" in stripped_line and "{" in stripped_line:
            inside_fpga_pr = True
            current_fpga_pr = [filtered_lines[i]]
            i += 1
            continue
        elif inside_fpga_pr:
            current_fpga_pr.append(filtered_lines[i])
            if "};" in stripped_line:
                fpga_pr_nodes.append(current_fpga_pr)
                inside_fpga_pr = False
            i += 1
            continue

        # Handle Misc Clock nodes
        if "misc_clk_" in stripped_line and "{" in stripped_line:
            inside_misc_clk = True
            current_misc_clk = [filtered_lines[i]]
            i += 1
            continue
        elif inside_misc_clk:
            current_misc_clk.append(filtered_lines[i])
            if "};" in stripped_line:
                misc_clk_nodes.append(current_misc_clk)
                inside_misc_clk = False
            i += 1
            continue

        # Handle Clocking nodes
        if "clocking" in stripped_line and "{" in stripped_line:
            inside_clocking = True
            current_clocking = [filtered_lines[i]]
            i += 1
            continue
        elif inside_clocking:
            current_clocking.append(filtered_lines[i])
            if "};" in stripped_line:
                clocking_nodes.append(current_clocking)
                inside_clocking = False
            i += 1
            continue

        # Handle afi nodes
        if "afi" in stripped_line and "{" in stripped_line:
            inside_afi = True
            current_afi = [filtered_lines[i]]
            i += 1
            continue
        elif inside_afi:
            current_afi.append(filtered_lines[i])
            if "};" in stripped_line:
                afi_nodes.append(current_afi)
                inside_afi = False
            i += 1
            continue

        # Store remaining amba nodes
        if amba_started:
            amba_block.append(filtered_lines[i])
        else:
            modified_lines.append(filtered_lines[i])

        i += 1

    # Append extracted FPGA_PR, misc_clk, and clocking nodes separately
    fpga_block = []
    if platform in ["ZynqMP", "Zynq"]:
        fpga_block.append("&fpga_full {\n")
    else:
        fpga_block.append("&fpga {\n")

    # Add firmware-name if present
    if platform not in ["ZynqMP", "Zynq"] and config == "dfx":
        firmware_name = "\t\texternal-fpga-config;\n"

    if config == "external-fpga-config":
        firmware_name = "\t\texternal-fpga-config;\n"

    if firmware_override:
        firmware_name = f'\t\tfirmware-name = "{firmware_override}";\n'

    if firmware_name:
        fpga_block.append(firmware_name)

    for pr_node in fpga_pr_nodes:
        fpga_block.extend(pr_node)
    for misc_node in misc_clk_nodes:
        fpga_block.extend(misc_node)
    for clocking_node in clocking_nodes:
        fpga_block.extend(clocking_node)
    for afi_node in afi_nodes:
        fpga_block.extend(afi_node)

    fpga_block.append("};\n\n")

    # Assemble final content
    final_lines = []
    final_lines.append("/dts-v1/;\n/plugin/;\n")
    final_lines.extend(fpga_block)  # FPGA-related nodes first
    final_lines.extend(modified_lines)  # General modifications
    if amba_block:
        final_lines.append("&amba {\n")  # Correct naming
        final_lines.extend(amba_block[1:])  # Avoid duplicate `&amba {`
        final_lines.append("};\n")

    # Write output file
    output_file = os.path.join(sdt.outdir, 'pl.dtsi')
    with open(output_file, "w") as outfile:
        for line in final_lines[:-1]:  # Exclude last line
            line = line[1:] if line.startswith("\t") else line  # Remove one leading tab
            line = line.replace("interrupt-parent = <&imux>;", "interrupt-parent = <&gic>;")  # Replace interrupt parent
            outfile.write(line)

    remove_status_okay(output_file)
    add_status_disabled_after_ipname(output_file, match_list)

    # Copy contents from pl.dtsi to pl.dtso in the same directory
    dtso_file = os.path.join(sdt.outdir, 'pl.dtso')
    with open(output_file, "r") as src, open(dtso_file, "w") as dst:
        dst.writelines(src.readlines())

    print("Overlay generation completed successfully!")
    return True




# /*
# * Copyright (c) 2019,2020,2021 Xilinx Inc. All rights reserved.
# *
# * Author:
# *     Ben Levinsky <ben.levinsky@xilinx.com>
# *     Izhar Shaikh <izhar.ameer.shaikh@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

from enum import IntEnum


class REQ_USAGE(IntEnum):
    REQ_NO_RESTRICTION = 0
    REQ_SHARED = 1
    REQ_NONSHARED = 2
    REQ_TIME_SHARED = 3


# if this bit combination is on for usage offset, the meaning is as described below
req_usage_message = "Device usage policies"
req_usage = {
    REQ_USAGE.REQ_NO_RESTRICTION:
    "device accessible from all subsystem",
    REQ_USAGE.REQ_SHARED:
    "device simultaneously shared between two or more subsystems",
    REQ_USAGE.REQ_NONSHARED:
    "device exclusively reserved by one subsystem, always",
    REQ_USAGE.REQ_TIME_SHARED:
    "device is time shared between two or more subsystems",
}

usage_mask = 0x3


def usage(flags):
    msg = "#    usage: "
    msg += req_usage[flags & usage_mask]
    return msg


class REGION_SECURITY(IntEnum):
    ACCESS_FROM_SECURE = 0
    ACCESS_FROM_NONSECURE = 1


req_security_message = "Device/Memory region security status requirement per TrustZone."
req_security = {
    REGION_SECURITY.ACCESS_FROM_SECURE:
    "Device/Memory region only allows access from secure masters",
    REGION_SECURITY.ACCESS_FROM_NONSECURE:
    "Device/Memory region allow both secure or non-secure masters",
}
security_mask = 0x4
security_offset = 0x2


def security(flags):
    msg = "#    security: "
    msg += req_security[(flags & security_mask) >> security_offset]
    return msg


class RDWR_POLICY(IntEnum):
    ALLOWED = 0
    NOT_ALLOWED = 1


# this map is only applicable for memory regions
req_rd_wr_message = "Read/Write access control policy"
req_rd_wr = {
    RDWR_POLICY.ALLOWED: "Transaction allowed",
    RDWR_POLICY.NOT_ALLOWED: "Transaction not Allowed",
}
rd_policy_mask = 0x8
rd_policy_offset = 0x3
wr_policy_mask = 0x10
wr_policy_offset = 0x4
rw_message = "Read/Write access control policy."


def read_policy(flags):
    msg = "#    read policy: "
    msg += req_rd_wr[(flags & rd_policy_mask) >> rd_policy_offset]
    return msg


def write_policy(flags):
    msg = "#    write policy: "
    msg += req_rd_wr[(flags & wr_policy_mask) >> wr_policy_offset]
    return msg


nsregn_check_mask = 0x20
nsregn_check_offset = 0x5


class NSREGN_POLICY(IntEnum):
    RELAXED = 0
    STRICT = 1


nsregn_message = "Non-secure memory region check type policy."
nsregn = {
    NSREGN_POLICY.RELAXED: "RELAXED",
    NSREGN_POLICY.STRICT: "STRICT",
}


def nsregn_policy(flags):
    msg = "#    Non-secure memory region check type policy: "
    msg += nsregn[(flags & nsregn_check_mask) >> nsregn_check_offset]
    return msg


capability_offset = 0x8
capability_mask = 0x7F00

cap_message = "capability: "


def capability_policy(flags):
    msg = "#    Capability policy: "
    msg += hex((flags & capability_mask) >> capability_offset)
    return msg


prealloc_offset = 6
prealloc_mask = (0x1 << 6)


class PREALLOC(IntEnum):
    NOT_REQUIRED = 0
    REQUIRED = 1


prealloc = {
    PREALLOC.NOT_REQUIRED: "prealloc not required",
    PREALLOC.REQUIRED: "prealloc required",
}

prealloc_message = "prealloc policy "


def prealloc_policy(flags):
    msg = "#    Preallocation policy: "
    msg += prealloc[(flags & prealloc_mask) >> prealloc_offset]
    return msg


def prealloc_detailed_policy(flags):
    msg = "#    Preallocation detailed: "
    caps = [
        "full access", "preserve context", "emit wake interrupts",
        "not usable", "secure access", "coherent access", "virtualized access"
    ]
    for index, s in enumerate(caps):
        match = ((0x1 << index) & flags) >> index
        if match == 1:
            msg += " " + s
    return msg


class Requirement:
    def __init__(self, subsystem, node, prealloc, capability, nsregn_policy,
                 read_policy, write_policy, security, usage):
        self.prealloc = prealloc
        self.capability = capability
        self.nsregn_policy = nsregn_policy
        self.read_policy = read_policy
        self.write_policy = write_policy
        self.security = security
        self.usage = usage
        self.subsystem = subsystem
        self.node = node


def mem_regn_node(node_id):
    return ((0x3F << 20) & node_id) == 0x300000


misc_devices = {
    "mailbox@ff320000": "PM_DEV_IPI_0",
    "mailbox@ff390000": "PM_DEV_IPI_1",
    "mailbox@ff310000": "PM_DEV_IPI_2",
    "mailbox@ff330000": "PM_DEV_IPI_3",
    "mailbox@ff340000": "PM_DEV_IPI_4",
    "mailbox@ff350000": "PM_DEV_IPI_5",
    "mailbox@ff360000": "PM_DEV_IPI_6",
    "watchdog@ff120000": "PM_DEV_SWDT_LPD",
}

xlnx_pm_mem_node_to_base = {
    "PM_DEV_OCM_0": 0xff960000,
    "PM_DEV_OCM_1": 0xff960000,
    "PM_DEV_OCM_2": 0xff960000,
    "PM_DEV_OCM_3": 0xff960000,
    "PM_DEV_TCM_0_A": 0xffe00000,
    "PM_DEV_TCM_0_B": 0xffe20000,
    "PM_DEV_TCM_1_A": 0xffe90000,
    "PM_DEV_TCM_1_B": 0xffeb0000,
}

xlnx_pm_devname_to_id = {
    "PM_DEV_PLD_0": 0x18700000,
    "PM_DEV_PMC_PROC": 0x18104001,
    "PM_DEV_PSM_PROC": 0x18108002,
    "PM_DEV_ACPU_0": 0x1810c003,
    "PM_DEV_ACPU_1": 0x1810c004,
    "PM_DEV_RPU0_0": 0x18110005,
    "PM_DEV_RPU0_1": 0x18110006,
    "PM_DEV_OCM_0": 0x18314007,
    "PM_DEV_OCM_1": 0x18314008,
    "PM_DEV_OCM_2": 0x18314009,
    "PM_DEV_OCM_3": 0x1831400a,
    "PM_DEV_TCM_0_A": 0x1831800b,
    "PM_DEV_TCM_0_B": 0x1831800c,
    "PM_DEV_TCM_1_A": 0x1831800d,
    "PM_DEV_TCM_1_B": 0x1831800e,
    "PM_DEV_L2_BANK_0": 0x1831c00f,
    "PM_DEV_DDR_0": 0x18320010,
    "PM_DEV_USB_0": 0x18224018,
    "PM_DEV_GEM_0": 0x18224019,
    "PM_DEV_GEM_1": 0x1822401a,
    "PM_DEV_SPI_0": 0x1822401b,
    "PM_DEV_SPI_1": 0x1822401c,
    "PM_DEV_I2C_0": 0x1822401d,
    "PM_DEV_I2C_1": 0x1822401e,
    "PM_DEV_CAN_FD_0": 0x1822401f,
    "PM_DEV_CAN_FD_1": 0x18224020,
    "PM_DEV_UART_0": 0x18224021,
    "PM_DEV_UART_1": 0x18224022,
    "PM_DEV_GPIO": 0x18224023,
    "PM_DEV_TTC_0": 0x18224024,
    "PM_DEV_TTC_1": 0x18224025,
    "PM_DEV_TTC_2": 0x18224026,
    "PM_DEV_TTC_3": 0x18224027,
    "PM_DEV_SWDT_LPD": 0x18224028,
    "PM_DEV_SWDT_FPD": 0x18224029,
    "PM_DEV_OSPI": 0x1822402a,
    "PM_DEV_QSPI": 0x1822402b,
    "PM_DEV_GPIO_PMC": 0x1822402c,
    "PM_DEV_I2C_PMC": 0x1822402d,
    "PM_DEV_SDIO_0": 0x1822402e,
    "PM_DEV_SDIO_1": 0x1822402f,
    "PM_DEV_RTC": 0x18224034,
    "PM_DEV_ADMA_0": 0x18224035,
    "PM_DEV_ADMA_1": 0x18224036,
    "PM_DEV_ADMA_2": 0x18224037,
    "PM_DEV_ADMA_3": 0x18224038,
    "PM_DEV_ADMA_4": 0x18224039,
    "PM_DEV_ADMA_5": 0x1822403a,
    "PM_DEV_ADMA_6": 0x1822403b,
    "PM_DEV_ADMA_7": 0x1822403c,
    "PM_DEV_IPI_0": 0x1822403d,
    "PM_DEV_IPI_1": 0x1822403e,
    "PM_DEV_IPI_2": 0x1822403f,
    "PM_DEV_IPI_3": 0x18224040,
    "PM_DEV_IPI_4": 0x18224041,
    "PM_DEV_IPI_5": 0x18224042,
    "PM_DEV_IPI_6": 0x18224043,
    "PM_DEV_SOC": 0x18428044,
    "PM_DEV_DDRMC_0": 0x18520045,
    "PM_DEV_DDRMC_1": 0x18520046,
    "PM_DEV_DDRMC_2": 0x18520047,
    "PM_DEV_DDRMC_3": 0x18520048,
    "PM_DEV_GT_0": 0x1862c049,
    "PM_DEV_GT_1": 0x1862c04a,
    "PM_DEV_GT_2": 0x1862c04b,
    "PM_DEV_GT_3": 0x1862c04c,
    "PM_DEV_GT_4": 0x1862c04d,
    "PM_DEV_GT_5": 0x1862c04e,
    "PM_DEV_GT_6": 0x1862c04f,
    "PM_DEV_GT_7": 0x1862c050,
    "PM_DEV_GT_8": 0x1862c051,
    "PM_DEV_GT_9": 0x1862c052,
    "PM_DEV_GT_10": 0x1862c053,
    "PM_DEV_EFUSE_CACHE": 0x18330054,
    "PM_DEV_AMS_ROOT": 0x18224055,
    "PM_DEV_AIE": 0x18224072,
    "PM_DEV_IPI_PMC": 0x18224073,
}

xlnx_pm_devid_to_name = {
    0x18700000: 'PM_DEV_PLD_0',
    0x18104001: 'PM_DEV_PMC_PROC',
    0x18108002: 'PM_DEV_PSM_PROC',
    0x1810c003: 'PM_DEV_ACPU_0',
    0x1810c004: 'PM_DEV_ACPU_1',
    0x18110005: 'PM_DEV_RPU0_0',
    0x18110006: 'PM_DEV_RPU0_1',
    0x18314007: 'PM_DEV_OCM_0',
    0x18314008: 'PM_DEV_OCM_1',
    0x18314009: 'PM_DEV_OCM_2',
    0x1831400a: 'PM_DEV_OCM_3',
    0x1831800b: 'PM_DEV_TCM_0_A',
    0x1831800c: 'PM_DEV_TCM_0_B',
    0x1831800d: 'PM_DEV_TCM_1_A',
    0x1831800e: 'PM_DEV_TCM_1_B',
    0x1831c00f: 'PM_DEV_L2_BANK_0',
    0x18320010: 'PM_DEV_DDR_0',
    0x18224018: 'PM_DEV_USB_0',
    0x18224019: 'PM_DEV_GEM_0',
    0x1822401a: 'PM_DEV_GEM_1',
    0x1822401b: 'PM_DEV_SPI_0',
    0x1822401c: 'PM_DEV_SPI_1',
    0x1822401d: 'PM_DEV_I2C_0',
    0x1822401e: 'PM_DEV_I2C_1',
    0x1822401f: 'PM_DEV_CAN_FD_0',
    0x18224020: 'PM_DEV_CAN_FD_1',
    0x18224021: 'PM_DEV_UART_0',
    0x18224022: 'PM_DEV_UART_1',
    0x18224023: 'PM_DEV_GPIO',
    0x18224024: 'PM_DEV_TTC_0',
    0x18224025: 'PM_DEV_TTC_1',
    0x18224026: 'PM_DEV_TTC_2',
    0x18224027: 'PM_DEV_TTC_3',
    0x18224028: 'PM_DEV_SWDT_LPD',
    0x18224029: 'PM_DEV_SWDT_FPD',
    0x1822402a: 'PM_DEV_OSPI',
    0x1822402b: 'PM_DEV_QSPI',
    0x1822402c: 'PM_DEV_GPIO_PMC',
    0x1822402d: 'PM_DEV_I2C_PMC',
    0x1822402e: 'PM_DEV_SDIO_0',
    0x1822402f: 'PM_DEV_SDIO_1',
    0x18224034: 'PM_DEV_RTC',
    0x18224035: 'PM_DEV_ADMA_0',
    0x18224036: 'PM_DEV_ADMA_1',
    0x18224037: 'PM_DEV_ADMA_2',
    0x18224038: 'PM_DEV_ADMA_3',
    0x18224039: 'PM_DEV_ADMA_4',
    0x1822403a: 'PM_DEV_ADMA_5',
    0x1822403b: 'PM_DEV_ADMA_6',
    0x1822403c: 'PM_DEV_ADMA_7',
    0x1822403d: 'PM_DEV_IPI_0',
    0x1822403e: 'PM_DEV_IPI_1',
    0x1822403f: 'PM_DEV_IPI_2',
    0x18224040: 'PM_DEV_IPI_3',
    0x18224041: 'PM_DEV_IPI_4',
    0x18224042: 'PM_DEV_IPI_5',
    0x18224043: 'PM_DEV_IPI_6',
    0x18428044: 'PM_DEV_SOC',
    0x18520045: 'PM_DEV_DDRMC_0',
    0x18520046: 'PM_DEV_DDRMC_1',
    0x18520047: 'PM_DEV_DDRMC_2',
    0x18520048: 'PM_DEV_DDRMC_3',
    0x1862c049: 'PM_DEV_GT_0',
    0x1862c04a: 'PM_DEV_GT_1',
    0x1862c04b: 'PM_DEV_GT_2',
    0x1862c04c: 'PM_DEV_GT_3',
    0x1862c04d: 'PM_DEV_GT_4',
    0x1862c04e: 'PM_DEV_GT_5',
    0x1862c04f: 'PM_DEV_GT_6',
    0x1862c050: 'PM_DEV_GT_7',
    0x1862c051: 'PM_DEV_GT_8',
    0x1862c052: 'PM_DEV_GT_9',
    0x1862c053: 'PM_DEV_GT_10',
    0x18330054: 'PM_DEV_EFUSE_CACHE',
    0x18224055: 'PM_DEV_AMS_ROOT',
    0x18224072: 'PM_DEV_AIE',
    0x18224073: 'PM_DEV_IPI_PMC',
}



#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import getopt
import os
import sys
import atexit
from pathlib import Path
import configparser
import re

from lopper import LopperSDT

from lopper.log import _warning, _info, _error, _debug
import logging
from lopper import lopper_directory

global device_tree
device_tree = None

def at_exit_cleanup():
    if device_tree:
        device_tree.cleanup()
    else:
        pass

with open(Path(__file__).parent / 'VERSION', 'r') as f:
    LOPPER_VERSION = f.read().strip()


def parse_schema_argument(schema_arg):
    """Parse schema argument to determine action and output path.

    Returns: (action, target) tuple where:
        action is one of: "none", "learn", "learn_dump", "load"
        target is: None, output path, or schema path
    """
    if not schema_arg:
        return ("learn", None)

    if schema_arg == "none":
        return ("none", None)
    elif schema_arg == "learn":
        return ("learn", None)
    elif schema_arg.startswith("learn:"):
        output_path = schema_arg[6:]  # Remove "learn:"
        if not output_path:
            _error("schema output path cannot be empty after 'learn:'", also_exit=1)
        return ("learn_dump", output_path)
    else:
        # Assume it's a path to an existing schema
        return ("load", schema_arg)

def usage():
    prog = "lopper"
    print(f'Usage: {prog} [OPTION] <system device tree> [<output file>]...')
    print('  -v, --verbose       enable verbose/debug processing (specify more than once for more verbosity)')
    print('  -t, --target        indicate the starting domain for processing (i.e. chosen node or domain label)' )
    print('    , --dryrun        run all processing, but don\'t write any output files' )
    print('  -d, --dump          dump a dtb as dts source' )
    print('  -i, --input         process supplied input device tree description')
    print('  -I, --input-dirs    colon separated list of directories to search for input files (any type)')
    print('                      input directories can also be set by environment variable LOPPER_INPUT_DIRS')
    print('  -a, --assist        load specified python assist (for node or output processing)' )
    print('  -A, --assist-paths  colon separated lists of paths to search for assist loading' )
    print('    , --enhanced      when writing output files, do enhanced processing (this includes phandle replacement, comments, etc' )
    print('    . --auto          automatically run any eligible assists (via -a) or lops (embedded)' )
    print('    , --permissive    do not enforce fully validated properties (phandles, etc)' )
    print('    , -W              enable a warning: '  )
    print('                          invalid_phandle (warn on invalid phandle references)' )
    print('                          duplicate_phandle (warn on duplicate phandle values)' )
    print('                          phandle_change (warn when phandle value changes)' )
    print('                          memory_cells (validate #address-cells, #size-cells)' )
    print('                          memory_reg (validate reg property format)' )
    print('                          memory_overlap (detect reserved-memory overlaps)' )
    print('                          reserved_bounds (reserved-memory in domain bounds)' )
    print('                          domain_overlap (overlaps within a domain)' )
    print('                          cross_domain_overlap (overlaps between domains)' )
    print('                          memory_all (enable all memory checks)' )
    print('                          schema_forbidden_props (check forbidden properties)' )
    print('                          schema_required_props (check required properties)' )
    print('                          schema_prop_values (check property values)' )
    print('                          schema_mutex_props (check mutually exclusive props)' )
    print('                          schema_reserved_memory (reserved-memory checks)' )
    print('                          schema_all (enable all schema checks)' )
    print('                          all (enable all warnings)' )
    print('    , --memmap        output file for memory map visualization (use - for stdout)' )
    print('    , --symbols       generate (and maintain) the __symbols__ node during processing' )
    print('  -o, --output        output file')
    print('    , --overlay       Allow input files (dts or yaml) to overlay system device tree nodes' )
    print('  -x. --xlate         run automatic translations on nodes for indicated input types (yaml,dts)' )
    print('    , --no-libfdt     don\'t use dtc/libfdt for parsing/compiling device trees' )
    print('  -f, --force         force overwrite output file(s)')
    print('    , --werror        treat warnings as errors' )
    print('    , --emit-overlay-sidecar  write a <out>.overlays.yaml sigil YAML alongside the output DTS' )
    print('    , --emit-overlay-dtso     write per-condition <out>.<cond>.dtso overlay files alongside the output DTS' )
    print('    , --emit-embedded-overlays  re-embed /__lopper-overlays__ even when loaded from a prior DTS embed (use for >2-pass workflows)' )
    print('  -S, --save-temps    don\'t remove temporary files' )
    print('    , --cfgfile       specify a lopper configuration file to use (configparser format) ' )
    print('    , --cfgval        specify a configuration value to use (in configparser section format). Can be specified multiple times' )
    print('    , --schema        one of: "path to a dts schema", "learn" or "none" ')
    print('  -h, --help          display this help and exit')
    print('  -O, --outdir        directory to use for output files')
    print('    , --server        after processing, start a server for ReST API calls')
    print('    , --version       output the version and exit')
    print('')

def main():
    debug = False
    sdt = None
    verbose = 0
    output = ""
    inputfiles = []
    force = False
    dump_dtb = False
    dryrun = False
    target_domain = ""
    cmdline_assists = []
    werror = False
    save_temps = False
    enhanced_print = False
    outdir="./"
    load_paths = []
    input_dirs = []
    server = False
    auto_run = False
    permissive = False
    libfdt = True
    xlate = []
    overlay = False
    config_file = None
    config_vals = {}
    symbols = False
    warnings = []
    usage_flag = False
    schema = None
    memmap_file = None
    cpumap_file = None
    cpumap_expand = False
    overlay_emit = set()

    try:
        opts, args = getopt.getopt(sys.argv[1:], "I:W:A:t:dfvdhi:o:a:SO:D:x:",
                                   [ "debug=", "assist-paths=", "outdir", "enhanced",
                                     "schema=", "save-temps", "version", "werror","target=", "dump",
                                     "force","verbose","help","input=","output=","dryrun",
                                     "assist=","server", "auto", "permissive", 'symbols', "xlate=",
                                     "no-libfdt", "overlay", "cfgfile=", "cfgval=", "input-dirs",
                                     "memmap=", "cpumap=", "cpumap-expand",
                                     "emit-overlay-sidecar", "emit-overlay-dtso",
                                     "emit-embedded-overlays"] )
    except getopt.GetoptError as err:
        _error(f"{err}")
        usage()
        sys.exit(2)

    if opts == [] and args == []:
        usage()
        sys.exit(1)

    for o, a in opts:
        if o in ('-v', "--verbose"):
            verbose = verbose + 1
        elif o in ('-d', "--dump"):
            dump_dtb = True
        elif o in ('-f', "--force"):
            force = True
        elif o in ('-h', '--help'):
            # usage()
            usage_flag = True
        elif o in ('-i', '--input'):
            inputfiles.append(a)
        elif o in ('-a', '--assist'):
            cmdline_assists.append(a)
        elif o in ('-A', '--assist-path'):
            load_paths += a.split(":")
        elif o in ('-I', '--input-dirs'):
            input_dirs += a.split(":")
        elif o in ('-O', '--outdir'):
            outdir = a
        elif o in ('-D', '--debug'):
            debug = a
        elif o in ('-t', '--target'):
            target_domain = a
        elif o in ('-o', '--output'):
            output = a
        elif o in ('--dryrun'):
            dryrun=True
        elif o in ('--werror'):
            werror=True
        elif o in ('--server'):
            server=True
        elif o in ('--schema'):
            schema = a
        elif o in ('-S', '--save-temps' ):
            save_temps=True
        elif o in ('--no-libfdt' ):
            libfdt=False
        elif o in ('--enhanced' ):
            enhanced_print = True
        elif o in ('--auto' ):
            auto_run = True
        elif o in ('--permissive' ):
            permissive = True
        elif o in ('--symbols' ):
            symbols = True
        elif o in ('--overlay' ):
            overlay = True
        elif o in ('--cfgfile' ):
            config_file = a
        elif o in ('--cfgval' ):
            config_vals[a] = a
        elif o in ('-x', '--xlate'):
            xlate.append(a)
        elif o in ('-W'):
            # warning processing
            warnings.append(a)
        elif o in ('--memmap'):
            memmap_file = a
        elif o in ('--cpumap'):
            cpumap_file = a
        elif o in ('--cpumap-expand'):
            cpumap_expand = True
        elif o in ('--emit-overlay-sidecar'):
            overlay_emit.add('sidecar')
        elif o in ('--emit-overlay-dtso'):
            overlay_emit.add('dtso')
        elif o in ('--emit-embedded-overlays'):
            overlay_emit.add('embedded')
        elif o in ('--version'):
            print( f"{LOPPER_VERSION}" )
            sys.exit(0)
        else:
            assert False, "unhandled option"


    # We split the options into two groups:
    #    1) options after -- on the command line
    #    2) options before the -- on the command line
    #
    # We can't just reply on getopt processing, since it will stop
    # handling arguments above when the firt unrecognized non dashed
    # option is found or when "--" is found as a delimeter.
    #
    # That's all fine, but it also doesn't tell us WHY it stopped
    # handling the options (-- or a non-dashed opttion), which means
    # we can't easily tell if a subcommand was being run, or it is a
    # system device tree that was being passed.
    #
    # So we double check against argv and split into the two parts
    # anything remaining before the dash could be a SDT, everything
    # after is for modules/commands
    #
    option_args = []
    non_option_args = []
    if '--' in sys.argv:
        double_dash_index = sys.argv.index('--', 1)
        # All arguments after '--' are not options
        option_args_possible = sys.argv[1:double_dash_index]
        non_option_args = sys.argv[double_dash_index + 1:]

        # Separate only unprocessed arguments
        option_args = [
            arg for arg in option_args_possible
            if arg not in [opt for opt, val in opts] and arg not in [val for opt, val in opts]
        ]

        # print( f"getopt remaining args: {args} option_args: {option_args} non_option_args: {non_option_args}" )
    else:
        option_args = args

    # any args should be <system device tree> <output file>
    module_name = ""
    module_args = {}
    module_args_found = False
    for idx, item in enumerate(option_args):
        # validate that the system device tree file exists
        if idx == 0:
            sdt = item
            sdt_file = Path(sdt)
            try:
                my_abs_path = sdt_file.resolve(strict=True)
            except FileNotFoundError:
                _error(f"system device tree {sdt} does not exist", also_exit=1)

        if idx == 1:
            if output:
                _error("output was already provided via -o")
                usage()
                sys.exit(1)
            else:
                output = item
                output_file = Path(output)
                if output_file.exists():
                    if not force:
                        _error(f"output file {output} exists, and -f was not passed", also_exit=1)

    # these are options that followed -- on the original command line
    for idx, item in enumerate(non_option_args):
        # check for chained modules "--"
        if not item == "--":
            if not module_name:
                module_name = item
                cmdline_assists.append( item )
                module_args[module_name] = []
            else:
                module_args[module_name].append( item )
        else:
            if module_name:
                # another module, clear the name to trigger a re-start of the
                # processing
                module_name = ""

    if module_name:
        _debug(f"modules found: {list(module_args.keys())}")
        _debug(f"         args: {module_args}")

    # was --help passed ?
    if usage_flag:
        if not module_name:
            usage()
        else:
            # a module name was found, let's pass this onto it
            pass

    if not usage_flag and not sdt:
        # if a module was found, pass along everything to it
        if not module_name:
            _error("no system device tree was supplied")
            usage()
            sys.exit(1)

    if not libfdt:
        import lopper.dt
        lopper.lopper_type(lopper.dt.LopperDT)
    else:
        import lopper.fdt
        lopper.lopper_type(lopper.fdt.LopperFDT)

    if outdir != "./":
        op = Path( outdir )
        try:
            op.resolve(True)
        except Exception:
            _error(f"output directory \"{outdir}\" does not exist", also_exit=1)

    # Not indicated in the help message, but we combine all the search
    # directories + environment variables. Assists, lops and dts files
    # can be found in any input directory type, but for the purpose of
    # documentation, they are considered separate for now.
    env_paths = (os.environ.get('LOPPER_INPUT_DIRS') or "").split(":")
    all_search_paths_as_passed = load_paths + input_dirs + env_paths
    all_search_paths = []
    for p in all_search_paths_as_passed:
        p_path = Path(p)
        abs_p_path = p_path.resolve()
        if not abs_p_path.exists():
            lopper.log._debug( f"input search directory {p_path} not found" )
        else:
            all_search_paths.append( abs_p_path.as_posix() )

    # check that the input files (passed via -i) exist
    for i in inputfiles:
        valid_ifile_types = [ ".json", ".dtsi", ".dtb", ".dts", ".yaml" ]
        itype = lopper.Lopper.input_file_type(i)
        if not itype in valid_ifile_types:
            _error("unrecognized input file type passed", also_exit=1)

    # config file handling
    config = configparser.ConfigParser()
    if not config_file:
        config_file = f"{lopper_directory}/lopper.ini"

    inf = Path(config_file)
    if not inf.exists():
        _error(f"config file {config_file} does not exist", also_exit=1)

    config.read( inf.absolute() )

    if config_vals:
        # was there a ".", if so that's the section split marker
        for i,k in config_vals.items():
            config_sections = k.split( '.' )
            if len(config_sections) > 1:
                # we have sections
                config_option = config_sections[-1]
                config_option_name = config_option.split('=')[0]
                config_option_val = config_option.split('=')[-1]
                if config_option_name == config_option_val:
                    config_option_val = True

                for item in config_sections[:-1]:
                    try:
                        section = config[item]
                    except:
                        config[item] = {}

                    config[item][config_option_name] = str(config_option_val)

            else:
                # global section, not currently implemented
                pass

    if schema:
        action, target = parse_schema_argument(schema)

        if action == "none":
            schema = None
        elif action == "learn":
            schema = "learn"
        elif action == "learn_dump":
            # Check if output file exists
            if target != "-":  # Not stdout
                output_path = Path(target)
                if output_path.exists():
                    _error(f"schema output file {target} already exists. Please remove it or choose a different filename.", also_exit=1)
            schema = ("learn_dump", target)
        elif action == "load":
            schemaf = Path(target)
            if not schemaf.exists():
                _error(f"schema file {target} does not exist", also_exit=1)
            schema = target
    else:
        schema = "learn"

    # Track if -x/--xlate was used for legacy fallback
    xlate_fallback = False
    if xlate:
        # -x/--xlate is deprecated in favor of --auto with BitBake-style lop matching
        # Enable autorun mode so that input files are matched against lops like %.yaml.lop
        auto_run = True
        xlate_fallback = True

    if dump_dtb:
        lopper.Lopper.dtb_dts_export( sdt, verbose )
        sys.exit(0)

    device_tree = LopperSDT( sdt )

    atexit.register(at_exit_cleanup)

    lopper.log._init( __name__ )
    lopper.log._init( "___main.py__" )

    lopper.log.init( verbose )

    # set some flags before we process the tree.
    device_tree.dryrun = dryrun
    device_tree.verbose = verbose
    device_tree.werror = werror
    device_tree.output_file = output
    device_tree.cleanup_flag = True
    device_tree.save_temps = save_temps
    device_tree.enhanced = enhanced_print
    device_tree.outdir = outdir
    device_tree.target_domain = target_domain
    device_tree.load_paths = all_search_paths
    device_tree.permissive = permissive
    device_tree.merge = overlay
    device_tree.autorun = auto_run
    device_tree.config = config
    device_tree.symbols = symbols
    device_tree.warnings = warnings
    device_tree.schema = schema
    device_tree.memmap_file = memmap_file
    device_tree.cpumap_file = cpumap_file
    device_tree.overlay_emit = overlay_emit

    # Backwards compatibility: if lop-xlate-yaml.dts is explicitly passed,
    # remove it from the input list and enable auto-matching so %.yaml.lop
    # handles the actual YAML processing. This avoids incompatibility errors
    # since %.yaml.lop declares incompatibility with lop-xlate-yaml.dts.
    legacy_xlate_files = [f for f in inputfiles if os.path.basename(f) == "lop-xlate-yaml.dts"]
    for f in legacy_xlate_files:
        inputfiles.remove(f)
    if legacy_xlate_files and not auto_run:
        auto_run = True
        xlate_fallback = True
        device_tree.autorun = True

    if auto_run:
        # look for lops that match the pattern of the input
        # files, if so, queue them to run

        # note: this may cause duplicates, since all input files are searched
        #       and they may already be on the list. duplicates will be dealt
        #       with later.
        auto_assists = device_tree.find_any_matching_assists( inputfiles + [sdt],
                                                               xlate_fallback=xlate_fallback )
        inputfiles.extend( auto_assists )

    device_tree.setup( sdt, inputfiles, "", force, libfdt, config )
    device_tree.assists_setup( cmdline_assists )

    if auto_run:
        for a in cmdline_assists:
            try:
                assist_args = module_args[a]
            except:
                assist_args = []

            device_tree.assist_autorun_setup( a, assist_args )
    else:
        # "modules" are assists passed after -- on the command line call to
        # lopper.
        if module_args:
            # This sets the trigger node of "/", and makes it autorun
            for module_name in reversed(list(module_args.keys())):
                m_args = module_args[module_name]
                device_tree.assist_autorun_setup( module_name, m_args )

    if debug:
        if debug == "profile":
            import cProfile
            cProfile.runctx( 'device_tree.perform_lops()', globals(), locals() )

        elif re.search( r'\.py$', debug ):
            from importlib.machinery import SourceFileLoader

            for p in load_paths:
                if p not in sys.path:
                    sys.path.append( p )

            try:
                imported_test = SourceFileLoader( "debug", debug ).load_module()
                func_name_to_call = Path( debug ).stem
                func_to_call = getattr( imported_test, func_name_to_call )
                func_to_call( device_tree )
            except Exception as e:
                _error(f"{e}")
        else:
            try:
                # is it a python string ? try compiling and runnig it
                block = compile( debug, '<string>', 'exec' )
                eval( block )
            except Exception as e:
                _error(f"{e}")

        sys.exit(1)
    else:
        device_tree.perform_lops()

    # Run audit validation phases if any warnings are enabled
    if warnings:
        from lopper.audit.base import run_audit_phase, ValidationPhase
        # Run EARLY and POST_YAML phases (POST_PROCESSING runs after domain processing)
        error_count = run_audit_phase(ValidationPhase.EARLY, device_tree.tree, warnings, werror)
        error_count += run_audit_phase(ValidationPhase.POST_YAML, device_tree.tree, warnings, werror)
        # Run POST_PROCESSING phase (final consistency checks)
        error_count += run_audit_phase(ValidationPhase.POST_PROCESSING, device_tree.tree, warnings, werror)
        if error_count > 0 and werror:
            _error(f"audit validation failed with {error_count} error(s)", also_exit=1)

    # Generate memory map visualization if requested
    if memmap_file:
        from lopper.audit.memviz import render_memory_map
        title = target_domain if target_domain else "System"
        memmap_output = render_memory_map(device_tree.tree, title=title)
        if memmap_file == "-":
            print(memmap_output)
        else:
            with open(memmap_file, 'w') as f:
                f.write(memmap_output)
            _info(f"memory map written to {memmap_file}")

    if not dryrun:
        # write any changes to the FDT, before we do our write
        if device_tree.dts:
            lopper.Lopper.sync( device_tree.FDT, device_tree.tree.export() )
            device_tree.write( enhanced = device_tree.enhanced )
    else:
        _info(f"--dryrun was passed, output file {output} not written")

    if server:
        _info("starting WSGI server")

        try:
            import lopper.rest
            rest_support = True
        except Exception as e:
            _error(f"rest support is not loaded, check dependencies: {e}")
            rest_support = False

        if rest_support:
            lopper.rest.sdt = device_tree
            lopper.rest.app.run()  # run our Flask app

        sys.exit(1)

    device_tree.cleanup()


if __name__ == "__main__":
    main()



# /*
# * Copyright (c) 2022 - 2023 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Madhav Bhatt <madhav.bhatt@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

rpu0_as_power_management_master = True
rpu1_as_power_management_master = True
apu_as_power_management_master = True
rpu0_as_reset_management_master = True
rpu1_as_reset_management_master = True
apu_as_reset_management_master = True
rpu0_as_overlay_config_master = False
rpu1_as_overlay_config_master = False
apu_as_overlay_config_master = False

subsys_str = "PMU Firmware:PMU|Secure Subsystem:"

node_map = {
	"NODE_APU" : { "label" : "NODE_APU", "periph" : "psu_cortexa53_0", "type" : "processor" },
	"NODE_APU_0" : { "label" : "NODE_APU_0", "periph" : "psu_cortexa53_0", "type" : "processor" },
	"NODE_APU_1" : { "label" : "NODE_APU_1", "periph" : "psu_cortexa53_1", "type" : "processor" },
	"NODE_APU_2" : { "label" : "NODE_APU_2", "periph" : "psu_cortexa53_2", "type" : "processor" },
	"NODE_APU_3" : { "label" : "NODE_APU_3", "periph" : "psu_cortexa53_3", "type" : "processor" },
	"NODE_RPU" : { "label" : "NODE_RPU", "periph" : "psu_cortexr5_0", "type" : "processor" },
	"NODE_RPU_0" : { "label" : "NODE_RPU_0", "periph" : "psu_cortexr5_0", "type" : "processor" },
	"NODE_RPU_1" : { "label" : "NODE_RPU_1", "periph" : "psu_cortexr5_1", "type" : "processor" },
	"NODE_PLD" : { "label" : "NODE_PLD", "periph" : "NA", "type" : "power" },
	"NODE_FPD" : { "label" : "NODE_FPD", "periph" : "NA", "type" : "power" },
	"NODE_OCM_BANK_0" : { "label" : "NODE_OCM_BANK_0", "periph" : "psu_ocm_0", "type" : "memory", "base_addr" : [0xfffc0000]},
	"NODE_OCM_BANK_1" : { "label" : "NODE_OCM_BANK_1", "periph" : "psu_ocm_1", "type" : "memory", "base_addr" : [0xfffd0000]},
	"NODE_OCM_BANK_2" : { "label" : "NODE_OCM_BANK_2", "periph" : "psu_ocm_2", "type" : "memory", "base_addr" : [0xfffe0000]},
	"NODE_OCM_BANK_3" : { "label" : "NODE_OCM_BANK_3", "periph" : "psu_ocm_3", "type" : "memory", "base_addr" : [0xffff0000]},
	"NODE_TCM_0_A" : { "label" : "NODE_TCM_0_A", "periph" : "psu_r5_0_atcm_global", "type" : "memory", "base_addr" : [0xffe00000] },
	"NODE_TCM_0_B" : { "label" : "NODE_TCM_0_B", "periph" : "psu_r5_0_btcm_global", "type" : "memory", "base_addr" : [0xffe20000] },
	"NODE_TCM_1_A" : { "label" : "NODE_TCM_1_A", "periph" : "psu_r5_1_atcm_global", "type" : "memory", "base_addr" : [0xffe90000] },
	"NODE_TCM_1_B" : { "label" : "NODE_TCM_1_B", "periph" : "psu_r5_1_btcm_global", "type" : "memory", "base_addr" : [0xffeb0000] },
	"NODE_L2" : { "label" : "NODE_L2", "periph" : "NA", "type" : "others" },
	"NODE_GPU_PP_0" : { "label" : "NODE_GPU_PP_0", "periph" : "psu_gpu", "type" : "slave", "base_addr" : [0xfd4b0000] },
	"NODE_GPU_PP_1" : { "label" : "NODE_GPU_PP_1", "periph" : "psu_gpu", "type" : "slave", "base_addr" : [0xfd4b0000] },
	"NODE_USB_0" : { "label" : "NODE_USB_0", "periph" : "psu_usb_0", "type" : "slave", "base_addr" : [0xff9d0000] },
	"NODE_USB_1" : { "label" : "NODE_USB_1", "periph" : "psu_usb_1", "type" : "slave", "base_addr" : [0xff9e0000] },
	"NODE_TTC_0" : { "label" : "NODE_TTC_0", "periph" : "psu_ttc_0", "type" : "slave", "base_addr" : [0xff110000] },
	"NODE_TTC_1" : { "label" : "NODE_TTC_1", "periph" : "psu_ttc_1", "type" : "slave", "base_addr" : [0xff120000] },
	"NODE_TTC_2" : { "label" : "NODE_TTC_2", "periph" : "psu_ttc_2", "type" : "slave", "base_addr" : [0xff130000 ] },
	"NODE_TTC_3" : { "label" : "NODE_TTC_3", "periph" : "psu_ttc_3", "type" : "slave", "base_addr" : [0xff140000] },
	"NODE_SATA" : { "label" : "NODE_SATA", "periph" : "psu_sata", "type" : "slave", "base_addr" : [0xfd0c0000] },
	"NODE_ETH_0" : { "label" : "NODE_ETH_0", "periph" : "psu_ethernet_0", "type" : "slave", "base_addr" : [0xff0b0000] },
	"NODE_ETH_1" : { "label" : "NODE_ETH_1", "periph" : "psu_ethernet_1", "type" : "slave", "base_addr" : [0xff0c0000] },
	"NODE_ETH_2" : { "label" : "NODE_ETH_2", "periph" : "psu_ethernet_2", "type" : "slave", "base_addr" : [0xff0d0000] },
	"NODE_ETH_3" : { "label" : "NODE_ETH_3", "periph" : "psu_ethernet_3", "type" : "slave", "base_addr" : [0xff0e0000] },
	"NODE_UART_0" : { "label" : "NODE_UART_0", "periph" : "psu_uart_0", "type" : "slave", "base_addr" : [0xff000000] },
	"NODE_UART_1" : { "label" : "NODE_UART_1", "periph" : "psu_uart_1", "type" : "slave", "base_addr" : [0xff010000] },
	"NODE_SPI_0" : { "label" : "NODE_SPI_0", "periph" : "psu_spi_0", "type" : "slave", "base_addr" : [0xff040000] },
	"NODE_SPI_1" : { "label" : "NODE_SPI_1", "periph" : "psu_spi_1", "type" : "slave", "base_addr" : [0xff050000] },
	"NODE_I2C_0" : { "label" : "NODE_I2C_0", "periph" : "psu_i2c_0", "type" : "slave", "base_addr" : [0xff020000] },
	"NODE_I2C_1" : { "label" : "NODE_I2C_1", "periph" : "psu_i2c_1", "type" : "slave", "base_addr" : [0xff030000] },
	"NODE_SD_0" : { "label" : "NODE_SD_0", "periph" : "psu_sd_0", "type" : "slave", "base_addr" : [0xff160000] },
	"NODE_SD_1" : { "label" : "NODE_SD_1", "periph" : "psu_sd_1", "type" : "slave", "base_addr" : [0xff170000] },
	"NODE_DP" : { "label" : "NODE_DP", "periph" : "psu_dp", "type" : "slave", "base_addr" : [0xfd4a0000] },
	"NODE_GDMA" : { "label" : "NODE_GDMA", "periph" : "psu_gdma_0", "type" : "slave", "base_addr" : [0xfd500000] },
	"NODE_ADMA" : { "label" : "NODE_ADMA", "periph" : "psu_adma_0", "type" : "slave", "base_addr" : [0xffa80000] },
	"NODE_NAND" : { "label" : "NODE_NAND", "periph" : "psu_nand_0", "type" : "slave", "base_addr" : [0xff100000] },
	"NODE_QSPI" : { "label" : "NODE_QSPI", "periph" : "psu_qspi_0", "type" : "slave", "base_addr" : [0xff0f0000] },
	"NODE_GPIO" : { "label" : "NODE_GPIO", "periph" : "psu_gpio_0", "type" : "slave", "base_addr" : [0xff0a0000] },
	"NODE_CAN_0" : { "label" : "NODE_CAN_0", "periph" : "psu_can_0", "type" : "slave", "base_addr" : [0xff060000] },
	"NODE_CAN_1" : { "label" : "NODE_CAN_1", "periph" : "psu_can_1", "type" : "slave", "base_addr" : [0xff070000] },
	"NODE_EXTERN" : { "label" : "NODE_EXTERN", "periph" : "NA", "type" : "others" },
	"NODE_DDR" : { "label" : "NODE_DDR", "periph" : "psu_ddr", "type" : "memory", "base_addr" : [0x100000,0x0] },
	"NODE_IPI_APU" : { "label" : "NODE_IPI_APU", "periph" : "NA", "type" : "ipi" },
	"NODE_IPI_RPU_0" : { "label" : "NODE_IPI_RPU_0", "periph" : "NA", "type" : "ipi" },
	"NODE_IPI_RPU_1" : { "label" : "NODE_IPI_RPU_1", "periph" : "NA", "type" : "ipi" },
	"NODE_GPU" : { "label" : "NODE_GPU", "periph" : "psu_gpu", "type" : "slave", "base_addr" : [0xfd4b0000] },
	"NODE_PCIE" : { "label" : "NODE_PCIE", "periph" : "psu_pcie", "type" : "slave", "base_addr" : [0xfd0e0000] },
	"NODE_PCAP" : { "label" : "NODE_PCAP", "periph" : "NA", "type" : "slave" },
	"NODE_RTC" : { "label" : "NODE_RTC", "periph" : "psu_rtc", "type" : "slave", "base_addr" : [0xffa60000] },
	"NODE_VCU" : { "label" : "NODE_VCU", "periph" : "vcu_0", "type" : "slave" },
	"NODE_PL" : { "label" : "NODE_PL", "periph" : "NA", "type" : "others" },
}

masters = {
               "psu_cortexa53_0" : {'name' : 'APU'},
               "psu_cortexr5_0" : {'name' : 'RPU0'},
               "psu_cortexr5_1" : {'name' : 'RPU1'}
          }

ocm_map = {
    "psu_ocm_0" : { "label" : "OCM_BANK_0", "base" : 0xFFFC0000, "high" : 0xFFFCFFFF },
    "psu_ocm_1" : { "label" : "OCM_BANK_1", "base" : 0xFFFD0000, "high" : 0xFFFDFFFF },
    "psu_ocm_2" : { "label" : "OCM_BANK_2", "base" : 0xFFFE0000, "high" : 0xFFFEFFFF },
    "psu_ocm_3" : { "label" : "OCM_BANK_3", "base" : 0xFFFF0000, "high" : 0xFFFFFFFF }
}

apu_prealloc_list = [
                        "NODE_DDR",
                        "NODE_L2",
                        "NODE_OCM_BANK_0",
                        "NODE_OCM_BANK_1",
                        "NODE_OCM_BANK_2",
                        "NODE_OCM_BANK_3",
                        "NODE_I2C_0",
                        "NODE_I2C_1",
                        "NODE_SD_1",
                        "NODE_QSPI",
                        "NODE_PL"
                    ]

rpu_0_prealloc_list = [
                        "NODE_TCM_0_A",
                        "NODE_TCM_0_B",
                        "NODE_TCM_1_A",
                        "NODE_TCM_1_B"
                     ]

rpu_0_prealloc_conditional_list = [
		        "NODE_DDR",
                        "NODE_OCM_BANK_0",
                        "NODE_OCM_BANK_1",
                        "NODE_OCM_BANK_2",
                        "NODE_OCM_BANK_3",
                        "NODE_I2C_0",
                        "NODE_I2C_1",
                        "NODE_SD_1",
                        "NODE_QSPI",
                        "NODE_PL",
                        "NODE_ADMA"
	            ]

rpu_1_prealloc_list = [
                        "NODE_TCM_1_A",
                        "NODE_TCM_1_B"
                     ]

hardcoded_proc_type = "psu_cortexa53_0"

power_node_list = [
                    "NODE_APU",
                    "NODE_RPU",
                    "NODE_FPD",
                    "NODE_PLD"
                  ]

power_perms = {
                "NODE_APU" : [ "psu_cortexr5_0", "psu_cortexr5_1" ],
                "NODE_RPU" : [ "psu_cortexa53_0", "psu_cortexr5_0", "psu_cortexr5_1" ],
                "NODE_FPD" : [ "psu_cortexr5_0", "psu_cortexr5_1" ],
                "NODE_PLD" : [ "psu_cortexa53_0", "psu_cortexr5_0", "psu_cortexr5_1" ],
}

reset_line_map = {
	"XILPM_RESET_PCIE_CFG" : { "label" : "XILPM_RESET_PCIE_CFG",  "type" : "rst_periph",  "node" : "NODE_PCIE" },
	"XILPM_RESET_PCIE_BRIDGE" : { "label" : "XILPM_RESET_PCIE_BRIDGE",  "type" : "rst_periph",  "node" : "NODE_PCIE" },
	"XILPM_RESET_PCIE_CTRL" : { "label" : "XILPM_RESET_PCIE_CTRL",  "type" : "rst_periph",  "node" : "NODE_PCIE" },
	"XILPM_RESET_DP" : { "label" : "XILPM_RESET_DP",  "type" : "rst_periph",  "node" : "NODE_DP" },
	"XILPM_RESET_SWDT_CRF" : { "label" : "XILPM_RESET_SWDT_CRF",  "type" : "normal" },
	"XILPM_RESET_AFI_FM5" : { "label" : "XILPM_RESET_AFI_FM5",  "type" : "normal" },
	"XILPM_RESET_AFI_FM4" : { "label" : "XILPM_RESET_AFI_FM4",  "type" : "normal" },
	"XILPM_RESET_AFI_FM3" : { "label" : "XILPM_RESET_AFI_FM3",  "type" : "normal" },
	"XILPM_RESET_AFI_FM2" : { "label" : "XILPM_RESET_AFI_FM2",  "type" : "normal" },
	"XILPM_RESET_AFI_FM1" : { "label" : "XILPM_RESET_AFI_FM1",  "type" : "normal" },
	"XILPM_RESET_AFI_FM0" : { "label" : "XILPM_RESET_AFI_FM0",  "type" : "normal" },
	"XILPM_RESET_GDMA" : { "label" : "XILPM_RESET_GDMA",  "type" : "rst_periph",  "node" : "NODE_GDMA" },
	"XILPM_RESET_GPU_PP1" : { "label" : "XILPM_RESET_GPU_PP1",  "type" : "rst_periph",  "node" : "NODE_GPU_PP_1" },
	"XILPM_RESET_GPU_PP0" : { "label" : "XILPM_RESET_GPU_PP0",  "type" : "rst_periph",  "node" : "NODE_GPU_PP_0" },
	"XILPM_RESET_GPU" : { "label" : "XILPM_RESET_GPU",  "type" : "rst_periph",  "node" : "NODE_GPU" },
	"XILPM_RESET_GT" : { "label" : "XILPM_RESET_GT",  "type" : "normal" },
	"XILPM_RESET_SATA" : { "label" : "XILPM_RESET_SATA",  "type" : "rst_periph",  "node" : "NODE_SATA" },
	"XILPM_RESET_ACPU3_PWRON" : { "label" : "XILPM_RESET_ACPU3_PWRON",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU2_PWRON" : { "label" : "XILPM_RESET_ACPU2_PWRON",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU1_PWRON" : { "label" : "XILPM_RESET_ACPU1_PWRON",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU0_PWRON" : { "label" : "XILPM_RESET_ACPU0_PWRON",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_APU_L2" : { "label" : "XILPM_RESET_APU_L2",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU3" : { "label" : "XILPM_RESET_ACPU3",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU2" : { "label" : "XILPM_RESET_ACPU2",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU1" : { "label" : "XILPM_RESET_ACPU1",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_ACPU0" : { "label" : "XILPM_RESET_ACPU0",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_DDR" : { "label" : "XILPM_RESET_DDR",  "type" : "rst_periph",  "node" : "NODE_DDR" },
	"XILPM_RESET_APM_FPD" : { "label" : "XILPM_RESET_APM_FPD",  "type" : "normal" },
	"XILPM_RESET_SOFT" : { "label" : "XILPM_RESET_SOFT",  "type" : "rst_shared" },
	"XILPM_RESET_GEM0" : { "label" : "XILPM_RESET_GEM0",  "type" : "rst_periph",  "node" : "NODE_ETH_0" },
	"XILPM_RESET_GEM1" : { "label" : "XILPM_RESET_GEM1",  "type" : "rst_periph",  "node" : "NODE_ETH_1" },
	"XILPM_RESET_GEM2" : { "label" : "XILPM_RESET_GEM2",  "type" : "rst_periph",  "node" : "NODE_ETH_2" },
	"XILPM_RESET_GEM3" : { "label" : "XILPM_RESET_GEM3",  "type" : "rst_periph",  "node" : "NODE_ETH_3" },
	"XILPM_RESET_QSPI" : { "label" : "XILPM_RESET_QSPI",  "type" : "rst_periph",  "node" : "NODE_QSPI" },
	"XILPM_RESET_UART0" : { "label" : "XILPM_RESET_UART0",  "type" : "rst_periph",  "node" : "NODE_UART_0" },
	"XILPM_RESET_UART1" : { "label" : "XILPM_RESET_UART1",  "type" : "rst_periph",  "node" : "NODE_UART_1" },
	"XILPM_RESET_SPI0" : { "label" : "XILPM_RESET_SPI0",  "type" : "rst_periph",  "node" : "NODE_SPI_0" },
	"XILPM_RESET_SPI1" : { "label" : "XILPM_RESET_SPI1",  "type" : "rst_periph",  "node" : "NODE_SPI_1" },
	"XILPM_RESET_SDIO0" : { "label" : "XILPM_RESET_SDIO0",  "type" : "normal" },
	"XILPM_RESET_SDIO1" : { "label" : "XILPM_RESET_SDIO1",  "type" : "normal" },
	"XILPM_RESET_CAN0" : { "label" : "XILPM_RESET_CAN0",  "type" : "rst_periph",  "node" : "NODE_CAN_0" },
	"XILPM_RESET_CAN1" : { "label" : "XILPM_RESET_CAN1",  "type" : "rst_periph",  "node" : "NODE_CAN_1" },
	"XILPM_RESET_I2C0" : { "label" : "XILPM_RESET_I2C0",  "type" : "rst_periph",  "node" : "NODE_I2C_0" },
	"XILPM_RESET_I2C1" : { "label" : "XILPM_RESET_I2C1",  "type" : "rst_periph",  "node" : "NODE_I2C_1" },
	"XILPM_RESET_TTC0" : { "label" : "XILPM_RESET_TTC0",  "type" : "rst_periph",  "node" : "NODE_TTC_0" },
	"XILPM_RESET_TTC1" : { "label" : "XILPM_RESET_TTC1",  "type" : "rst_periph",  "node" : "NODE_TTC_1" },
	"XILPM_RESET_TTC2" : { "label" : "XILPM_RESET_TTC2",  "type" : "rst_periph",  "node" : "NODE_TTC_2" },
	"XILPM_RESET_TTC3" : { "label" : "XILPM_RESET_TTC3",  "type" : "rst_periph",  "node" : "NODE_TTC_3" },
	"XILPM_RESET_SWDT_CRL" : { "label" : "XILPM_RESET_SWDT_CRL",  "type" : "normal" },
	"XILPM_RESET_NAND" : { "label" : "XILPM_RESET_NAND",  "type" : "rst_periph",  "node" : "NODE_NAND" },
	"XILPM_RESET_ADMA" : { "label" : "XILPM_RESET_ADMA",  "type" : "rst_periph",  "node" : "NODE_ADMA" },
	"XILPM_RESET_GPIO" : { "label" : "XILPM_RESET_GPIO",  "type" : "normal" },
	"XILPM_RESET_IOU_CC" : { "label" : "XILPM_RESET_IOU_CC",  "type" : "normal" },
	"XILPM_RESET_TIMESTAMP" : { "label" : "XILPM_RESET_TIMESTAMP",  "type" : "normal" },
	"XILPM_RESET_RPU_R50" : { "label" : "XILPM_RESET_RPU_R50",  "type" : "rst_proc", "proc" : "RPU_0" },
	"XILPM_RESET_RPU_R51" : { "label" : "XILPM_RESET_RPU_R51",  "type" : "rst_proc", "proc" : "RPU_1" },
	"XILPM_RESET_RPU_AMBA" : { "label" : "XILPM_RESET_RPU_AMBA",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_OCM" : { "label" : "XILPM_RESET_OCM",  "type" : "rst_periph",  "node" : "NODE_OCM_BANK_0" },
	"XILPM_RESET_RPU_PGE" : { "label" : "XILPM_RESET_RPU_PGE",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_USB0_CORERESET" : { "label" : "XILPM_RESET_USB0_CORERESET",  "type" : "rst_periph",  "node" : "NODE_USB_0" },
	"XILPM_RESET_USB1_CORERESET" : { "label" : "XILPM_RESET_USB1_CORERESET",  "type" : "rst_periph",  "node" : "NODE_USB_1" },
	"XILPM_RESET_USB0_HIBERRESET" : { "label" : "XILPM_RESET_USB0_HIBERRESET",  "type" : "rst_periph",  "node" : "NODE_USB_0" },
	"XILPM_RESET_USB1_HIBERRESET" : { "label" : "XILPM_RESET_USB1_HIBERRESET",  "type" : "rst_periph",  "node" : "NODE_USB_1" },
	"XILPM_RESET_USB0_APB" : { "label" : "XILPM_RESET_USB0_APB",  "type" : "rst_periph",  "node" : "NODE_USB_0" },
	"XILPM_RESET_USB1_APB" : { "label" : "XILPM_RESET_USB1_APB",  "type" : "rst_periph",  "node" : "NODE_USB_1" },
	"XILPM_RESET_IPI" : { "label" : "XILPM_RESET_IPI",  "type" : "rst_shared" },
	"XILPM_RESET_APM_LPD" : { "label" : "XILPM_RESET_APM_LPD",  "type" : "normal" },
	"XILPM_RESET_RTC" : { "label" : "XILPM_RESET_RTC",  "type" : "rst_periph",  "node" : "NODE_RTC" },
	"XILPM_RESET_SYSMON" : { "label" : "XILPM_RESET_SYSMON",  "type" : "NA" },
	"XILPM_RESET_AFI_FM6" : { "label" : "XILPM_RESET_AFI_FM6",  "type" : "normal" },
	"XILPM_RESET_LPD_SWDT" : { "label" : "XILPM_RESET_LPD_SWDT",  "type" : "normal" },
	"XILPM_RESET_FPD" : { "label" : "XILPM_RESET_FPD",  "type" : "rpu_only" },
	"XILPM_RESET_RPU_DBG1" : { "label" : "XILPM_RESET_RPU_DBG1",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_RPU_DBG0" : { "label" : "XILPM_RESET_RPU_DBG0",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_DBG_LPD" : { "label" : "XILPM_RESET_DBG_LPD",  "type" : "normal" },
	"XILPM_RESET_DBG_FPD" : { "label" : "XILPM_RESET_DBG_FPD",  "type" : "normal" },
	"XILPM_RESET_APLL" : { "label" : "XILPM_RESET_APLL",  "type" : "rst_proc", "proc" : "APU" },
	"XILPM_RESET_DPLL" : { "label" : "XILPM_RESET_DPLL",  "type" : "rst_shared" },
	"XILPM_RESET_VPLL" : { "label" : "XILPM_RESET_VPLL",  "type" : "rst_shared" },
	"XILPM_RESET_IOPLL" : { "label" : "XILPM_RESET_IOPLL",  "type" : "rst_shared" },
	"XILPM_RESET_RPLL" : { "label" : "XILPM_RESET_RPLL",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_GPO3_PL_0" : { "label" : "XILPM_RESET_GPO3_PL_0",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_1" : { "label" : "XILPM_RESET_GPO3_PL_1",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_2" : { "label" : "XILPM_RESET_GPO3_PL_2",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_3" : { "label" : "XILPM_RESET_GPO3_PL_3",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_4" : { "label" : "XILPM_RESET_GPO3_PL_4",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_5" : { "label" : "XILPM_RESET_GPO3_PL_5",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_6" : { "label" : "XILPM_RESET_GPO3_PL_6",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_7" : { "label" : "XILPM_RESET_GPO3_PL_7",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_8" : { "label" : "XILPM_RESET_GPO3_PL_8",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_9" : { "label" : "XILPM_RESET_GPO3_PL_9",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_10" : { "label" : "XILPM_RESET_GPO3_PL_10",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_11" : { "label" : "XILPM_RESET_GPO3_PL_11",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_12" : { "label" : "XILPM_RESET_GPO3_PL_12",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_13" : { "label" : "XILPM_RESET_GPO3_PL_13",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_14" : { "label" : "XILPM_RESET_GPO3_PL_14",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_15" : { "label" : "XILPM_RESET_GPO3_PL_15",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_16" : { "label" : "XILPM_RESET_GPO3_PL_16",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_17" : { "label" : "XILPM_RESET_GPO3_PL_17",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_18" : { "label" : "XILPM_RESET_GPO3_PL_18",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_19" : { "label" : "XILPM_RESET_GPO3_PL_19",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_20" : { "label" : "XILPM_RESET_GPO3_PL_20",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_21" : { "label" : "XILPM_RESET_GPO3_PL_21",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_22" : { "label" : "XILPM_RESET_GPO3_PL_22",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_23" : { "label" : "XILPM_RESET_GPO3_PL_23",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_24" : { "label" : "XILPM_RESET_GPO3_PL_24",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_25" : { "label" : "XILPM_RESET_GPO3_PL_25",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_26" : { "label" : "XILPM_RESET_GPO3_PL_26",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_27" : { "label" : "XILPM_RESET_GPO3_PL_27",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_28" : { "label" : "XILPM_RESET_GPO3_PL_28",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_29" : { "label" : "XILPM_RESET_GPO3_PL_29",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_30" : { "label" : "XILPM_RESET_GPO3_PL_30",  "type" : "normal" },
	"XILPM_RESET_GPO3_PL_31" : { "label" : "XILPM_RESET_GPO3_PL_31",  "type" : "normal" },
	"XILPM_RESET_RPU_LS" : { "label" : "XILPM_RESET_RPU_LS",  "type" : "rst_proc", "proc" : "RPU" },
	"XILPM_RESET_PS_ONLY" : { "label" : "XILPM_RESET_PS_ONLY",  "type" : "normal" },
	"XILPM_RESET_PL" : { "label" : "XILPM_RESET_PL",  "type" : "normal" },
	"XILPM_RESET_GPIO5_EMIO_92" : { "label" : "XILPM_RESET_GPIO5_EMIO_92",  "type" : "normal" },
	"XILPM_RESET_GPIO5_EMIO_93" : { "label" : "XILPM_RESET_GPIO5_EMIO_93",  "type" : "normal" },
	"XILPM_RESET_GPIO5_EMIO_94" : { "label" : "XILPM_RESET_GPIO5_EMIO_94",  "type" : "normal" },
	"XILPM_RESET_GPIO5_EMIO_95" : { "label" : "XILPM_RESET_GPIO5_EMIO_95",  "type" : "normal" },
}

gpo_nums = [2,3,4,5]




#/*
# * Copyright (c) 2024,2025,2026 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

"""
Lopper audit module - backwards compatibility shim.

This module has been restructured into a package (lopper/audit/).
This file provides backwards compatibility for code that imports
lopper.audit directly.

The actual implementation is in:
- lopper/audit/core.py: Phandle validation, basic tree checks
- lopper/audit/memory.py: Memory region validation
- lopper/audit/memviz.py: Memory map visualization
"""

# Import everything from the audit package
from lopper.audit import *
